import sys
import pandas as pd
import biogeme.database as db
import biogeme.biogeme as bio
import biogeme.models as models
import unittest
from biogeme.expressions import Beta, DefineVariable, Elem, Derive

pandas = pd.read_csv("swissmetro.dat",sep='\t')
database = db.Database("swissmetro",pandas)

# The Pandas data structure is available as database.data. Use all the
# Pandas functions to invesigate the database
#print(database.data.describe())

globals().update(database.variables)

# Removing some observations can be done directly using pandas.
#remove = (((database.data.PURPOSE != 1) & (database.data.PURPOSE != 3)) | (database.data.CHOICE == 0))
#database.data.drop(database.data[remove].index,inplace=True)

# Here we use the "biogeme" way for backward compatibility
exclude = (( PURPOSE != 1 ) * (  PURPOSE   !=  3  ) +  ( CHOICE == 0 )) > 0
database.remove(exclude)

ASC_TRAIN = Beta('ASC_TRAIN', -0.701188,None,None,0)
B_TIME = Beta('B_TIME', -1.27786,None,None,0)
B_COST = Beta('B_COST', -1.08379,None,None,0)
ASC_SM = Beta('ASC_SM', 0,None,None,0)
ASC_CAR = Beta('ASC_CAR', -0.154633,None,None,0)

SM_COST =  SM_CO   * (  GA   ==  0  ) 
TRAIN_COST =  TRAIN_CO   * (  GA   ==  0  )

TRAIN_TT_SCALED = TRAIN_TT  / 100.0
TRAIN_COST_SCALED =  TRAIN_COST / 100
SM_TT_SCALED = SM_TT / 100.0
SM_COST_SCALED = SM_COST / 100.0
CAR_TT_SCALED = CAR_TT / 100.0
CAR_CO_SCALED = CAR_CO / 100.0

V1 = ASC_TRAIN + \
     B_TIME * TRAIN_TT_SCALED + \
     B_COST * TRAIN_COST_SCALED
V2 = ASC_SM + \
     B_TIME * SM_TT_SCALED + \
     B_COST * SM_COST_SCALED
V3 = ASC_CAR + \
     B_TIME * CAR_TT_SCALED + \
     B_COST * CAR_CO_SCALED

# Associate utility functions with the numbering of alternatives
V = {1: V1,
     2: V2,
     3: V3}


# Associate the availability conditions with the alternatives
CAR_AV_SP =  DefineVariable('CAR_AV_SP',CAR_AV  * (  SP   !=  0  ),database)
TRAIN_AV_SP =  DefineVariable('TRAIN_AV_SP',TRAIN_AV  * (  SP   !=  0  ),database)

av = {1: TRAIN_AV_SP,
      2: SM_AV,
      3: CAR_AV_SP}

# The choice model is a logit, with availability conditions
prob1 = Elem({0:0,1:models.logit(V,av,1)},av[1])

# Elasticities can be computed. We illustrate below two
# formulas. Check in the output file that they produce the same
# result.

# First, the general definition of elasticities. This illustrates the
# use of the Derive expression, and can be used with any model,
# however complicated it is. Note the quotes in the Derive opertor.

genelas1 = Derive(prob1,'TRAIN_TT') * TRAIN_TT / prob1

# Second, the elasticity of logit models. See Ben-Akiva and Lerman for
# the formula

logitelas1 = TRAIN_AV_SP * (1.0 - prob1) * TRAIN_TT_SCALED * B_TIME

simulate = {'P1': prob1,
            'logit elas. 1':logitelas1,
            'generic elas. 1':genelas1}

class test_01simul(unittest.TestCase):
    def testSimulation(self):
        biogeme  = bio.BIOGEME(database,simulate)
        biogeme.modelName = "01logit_simul"
        results = biogeme.simulate()
        self.assertAlmostEqual(sum(results['P1']),907.9992101964821,2)
        self.assertAlmostEqual(sum(results['logit elas. 1']),-12673.838605478186,2)
        self.assertAlmostEqual(sum(results['generic elas. 1']),-12673.838605478186,2)
        


if __name__ == '__main__':
    unittest.main()





"""File 00factorAnalysis.py

Preliminary analysis of the indicators using factor analysis.

:author: Michel Bierlaire, EPFL
:date: Mon Sep  9 16:04:57 2019

"""
import pandas as pd
import numpy as np

# The following package can be installed using
# pip install factor_analyzer
# See https://github.com/EducationalTestingService/factor_analyzer
from factor_analyzer import FactorAnalyzer

# We first extract the columns containing the indicators

variables = ["Envir01",
        "Envir02",
	"Envir03",
	"Envir04",
	"Envir05",
	"Envir06",
	"Mobil01",
	"Mobil02",
	"Mobil03",
	"Mobil04",
	"Mobil05",
	"Mobil06",
	"Mobil07",
	"Mobil08",
	"Mobil09",
	"Mobil10",
	"Mobil11",
	"Mobil12",
	"Mobil13",
	"Mobil14",
	"Mobil15",
	"Mobil16",
	"Mobil17",
	"Mobil18",
	"Mobil19",
	"Mobil20",
	"Mobil21",
	"Mobil22",
	"Mobil23",
	"Mobil24",
	"Mobil25",
	"Mobil26",
	"Mobil27",
	"ResidCh01",
	"ResidCh02",
	"ResidCh03",
	"ResidCh04",
	"ResidCh05",
	"ResidCh06",
	"ResidCh07",
	"LifSty01",
	"LifSty02",
	"LifSty03",
	"LifSty04",
	"LifSty05",
	"LifSty06",
	"LifSty07",
	"LifSty08",
	"LifSty09",
	"LifSty10",
	"LifSty11",
	"LifSty12",
	"LifSty13",
	"LifSty14"]
indicators = pd.read_csv("optima.dat",sep='\t',usecols=variables)

# Negative values are missing values. 
indicators[indicators <= 0] = np.nan
indicators = indicators.dropna(axis = 0, how = 'any') 

# We perform the factor analysis
fa = FactorAnalyzer(rotation='varimax')
fa.fit(indicators)

# We obtain the factor loadings and label them
labeledResults = pd.DataFrame(fa.loadings_)
labeledResults.index = variables

# We keep only loadings that are 0.4 or higher, in absolute value. 
filter = (labeledResults <= 0.4) & (labeledResults >= -0.4)
labeledResults[filter] = ''
print(labeledResults)



"""
Working file for constrained optimization. Will be transferred to optimization.py when fimished and tested.

:author: Michel Bierlaire
:date: Mon Dec 23 16:36:03 2019

"""



import numpy as np
import biogeme.exceptions as excep
import scipy.optimize as sc
import scipy.linalg as la
import inspect as ip
import biogeme.messaging as msg

def project(x,bounds):
    """Project a point onto the feasible domain defined by the bounds.

    :param x: point to project
    :type x: numpy.array

    :param bounds: list of tuples (ell,u) containing the lower and upper bounds
          for each free parameter
    :type bounds: list(tuple)

    :return: projected point
    :rtype: numpy.array
    """

    if len(x) != len(bounds):
        raise excep.biogemeError(f"Incompatible size: {len(x)} and {len(bounds)}")

    y = np.asarray([np.minimum(bounds[i][1],np.maximum(bounds[i][0],x[i])) for i in range(len(x))])
    return y

def feasible(x, bounds):
    """ Check if point verifies the bound constraints

    :param x: point to project
    :type x: numpy.array

    :param bounds: list of tuples (ell,u) containing the lower and upper bounds
          for each variable
    :type bounds: list(tuple)

    :return: True if x is feasible, False otherwise.
    :rtype: bool
    """
    if len(x) != len(bounds):
        raise excep.biogemeError(f"Incompatible size: {len(x)} and {len(bounds)}")
    for i in range(len(x)):
        if x[i] < bounds[i][0]:
            return False
        if x[i] > bounds[i][1]:
            return False
    return True

def breakPoints(x, d, radius, bounds):
    """ Projects the direction d, starting from x, on the intersection of
the bound constraints, and the trust region of given radius.

    :param x: current point
    :type x: numpy.array

    :param d: search direction
    :type d: numpy.array

    :param radius: radius of the trust region
    :type radius: float

    :param bounds: list of tuples (ell,u) containing the lower and upper bounds
          for each free parameter
    :type bounds: list(tuple)

    :return: list of tuple (index, value), where index is the index of the variable, and value the value of the corresponding breakpoint.
    :rtype: list(tuple(int,float))

    """
    n = len(x)
    if len(d) != n:
        raise excep.biogemeError(f"Incompatible size: {n} and {len(d)}")
    if len(bounds) != n:
        raise excep.biogemeError(f"Incompatible size: {n} and {len(bounds)}")

    breakPoints = [np.minimum(bounds[i][1]-x[i],radius)/d[i] if d[i] > np.finfo(float).eps else np.maximum(bounds[i][0]-x[i],-radius)/d[i] if d[i] < -np.finfo(float).eps else 0 for i in range(n)]

    if any(b < 0 for b in breakPoints):
        raise excep.biogemeError("Infeasible point")

    return sorted(enumerate(breakPoints), key=lambda x:x[1])

def generalizedCauchyPoint(xk,gk,H,bounds,radius):
    """ Implementation of Step 2 of the Specific Algorithm by `Conn et al. (1988)`_. 

    .. _`Conn et al. (1988)`: https://www.ams.org/journals/mcom/1988-50-182/S0025-5718-1988-0929544-3/S0025-5718-1988-0929544-3.pdf

    :param xk: current point
    :type xk: numpy.array. Dimension n.

    :param gk: vector g involved in the quadratic model definition.
    :type gk: numpy.array. Dimension n.
    
    :param H: matrix H involved in the quadratic model definition.
    :type H: numpy.array. Dimension n x n.

    :param bounds: list of tuples (ell,u) containing the lower and upper bounds
          for each free parameter
    :type bounds: list(tuple)

    :param radius: radius of the trust region
    :type radius: float

    :return: generalized Cauchy point based on inexact line search.
    :rtype: numpy.array. Dimension n.

    """

    x = xk
    g = gk - H @ xk
    d = -gk
    
    bp = breakPoints(xk,d,radius,bounds)

    currentBreakPoint = 0
    while currentBreakPoint < len(bp) and bp[currentBreakPoint][1] == 0.0:
        d[bp[currentBreakPoint][0]] == 0.0
        currentBreakPoint += 1

    fprime = np.inner(gk,d)

    if fprime >= 0:
        return x

    fsecond = np.inner(d, H @ d)

    #for b in range(currentBreakPoint,len(bp)):
        
    
    tmin = 0
    tmax = np.finfo(np.float).max
    t = radius / np.inner(g,g)
    mx = model(x)
    logger.debug(f'x={x}')
    logger.debug(f'g={g}')
    while tmax - tmin > np.finfo(np.float).eps:
        logger.debug(f't={t}')
        logger.debug(f'x - t * g = {x - t * g}')
        p = project(x - t * g,bounds)
        logger.debug(f'p={p}')
        s = p - x
        logger.debug(f's={s}')
        gs = np.inner(g,s)
        logger.debug(f'g^Ts = {gs}')
        snorm = la.norm(s)
        logger.debug(f'snorm={snorm}')
        m = model(p)

        logger.debug(f'{m} <= {mx} + {klbs} * {gs} = {mx + klbs * gs}')
        logger.debug(f'')
        if snorm > radius or m > mx + kubs * gs:
            logger.debug(f'tmax={t}')
            tmax = t
        elif snorm < kfrd * radius and m < mx + klbs * gs:
            logger.debug(f'tmin={t}')
            tmin = t
        else:
            return p

        if tmax == np.finfo(np.float).max:
            t = 2 * t
        else:
            t = 0.5 * (tmin + tmax)
    
    logger.warning(f'Inexact line search is not converging')
    return x




"""File 04latentChoiceSeq.py

Choice model with the latent variable.
Mixture of logit.
Measurement equation for the indicators.
Sequential estimation.

:author: Michel Bierlaire, EPFL
:date: Tue Sep 10 08:13:18 2019

"""
import pandas as pd
import numpy as np
import biogeme.database as db
import biogeme.biogeme as bio
import biogeme.models as models
import biogeme.distributions as dist
import biogeme.results as res
from biogeme.expressions import Beta, DefineVariable, RandomVariable, exp, log, Integrate

# Read the data
df = pd.read_csv("optima.dat",sep='\t')
database = db.Database("optima",df)

# The following statement allows you to use the names of the variable
# as Python variable.
globals().update(database.variables)

# Exclude observations such that the chosen alternative is -1
exclude = (Choice == -1.0)
database.remove(exclude)


### Variables

ScaledIncome = DefineVariable('ScaledIncome',\
                              CalculatedIncome / 1000,database)
thresholds = [4,6,8,10]
ContIncome = models.piecewise(ScaledIncome,thresholds)
ContIncome_0_4000 = ContIncome[0]
ContIncome_4000_6000 = ContIncome[1]
ContIncome_6000_8000 = ContIncome[2]
ContIncome_8000_10000 = ContIncome[3]
ContIncome_10000_more = ContIncome[4]

# Definition of other variables
age_65_more = DefineVariable('age_65_more',age >= Numeric(65),database)
moreThanOneCar = DefineVariable('moreThanOneCar',NbCar > 1,database)
moreThanOneBike = DefineVariable('moreThanOneBike',NbBicy > 1,database)
individualHouse = DefineVariable('individualHouse',\
                                 HouseType == 1,database)
male = DefineVariable('male',Gender == 1,database)
haveChildren = DefineVariable('haveChildren',\
                              ((FamilSitu == 3)+(FamilSitu == 4)) > 0,database)
haveGA = DefineVariable('haveGA',GenAbST == 1,database)
highEducation = DefineVariable('highEducation', Education >= 6,database)

### Coefficients
# Read the estimates from the structural equation estimation
structResults = res.bioResults(pickleFile='02oneLatentOrdered.pickle')
structBetas = structResults.getBetaValues()

coef_intercept = structBetas['coef_intercept']
coef_age_65_more = structBetas['coef_age_65_more']
coef_haveGA = structBetas['coef_haveGA']
coef_ContIncome_0_4000 = structBetas['coef_ContIncome_0_4000']
coef_ContIncome_4000_6000 = structBetas['coef_ContIncome_4000_6000']
coef_ContIncome_6000_8000 = structBetas['coef_ContIncome_6000_8000']
coef_ContIncome_8000_10000 = structBetas['coef_ContIncome_8000_10000']
coef_ContIncome_10000_more = structBetas['coef_ContIncome_10000_more']
coef_moreThanOneCar = structBetas['coef_moreThanOneCar']
coef_moreThanOneBike = structBetas['coef_moreThanOneBike']
coef_individualHouse = structBetas['coef_individualHouse']
coef_male = structBetas['coef_male']
coef_haveChildren = structBetas['coef_haveChildren']
coef_highEducation = structBetas['coef_highEducation']

### Latent variable: structural equation

# Note that the expression must be on a single line. In order to 
# write it across several lines, each line must terminate with 
# the \ symbol

# Define a random parameter, normally distributed, designed to be used
# for numerical integration
omega = RandomVariable('omega')
density = dist.normalpdf(omega) 
sigma_s = Beta('sigma_s',1,-1000,1000,0)

CARLOVERS = \
            coef_intercept +\
            coef_age_65_more * age_65_more +\
            coef_ContIncome_0_4000 * ContIncome_0_4000 +\
            coef_ContIncome_4000_6000 * ContIncome_4000_6000 +\
            coef_ContIncome_6000_8000 * ContIncome_6000_8000 +\
            coef_ContIncome_8000_10000 * ContIncome_8000_10000 +\
            coef_ContIncome_10000_more * ContIncome_10000_more +\
            coef_moreThanOneCar * moreThanOneCar +\
            coef_moreThanOneBike * moreThanOneBike +\
            coef_individualHouse * individualHouse +\
            coef_male * male +\
            coef_haveChildren * haveChildren +\
            coef_haveGA * haveGA +\
            coef_highEducation * highEducation +\
            sigma_s * omega


# Choice model


ASC_CAR	 = Beta('ASC_CAR',0,-10000,10000,0)
ASC_PT	 = Beta('ASC_PT',0,-10000,10000,1)
ASC_SM	 = Beta('ASC_SM',0,-10000,10000,0)
BETA_COST_HWH = Beta('BETA_COST_HWH',0.0,-10000,10000,0 )
BETA_COST_OTHER = Beta('BETA_COST_OTHER',0.0,-10000,10000,0 )
BETA_DIST	 = Beta('BETA_DIST',0.0,-10000,10000,0)
BETA_TIME_CAR_REF = Beta('BETA_TIME_CAR_REF',0.0,-10000,0,0)
BETA_TIME_CAR_CL = Beta('BETA_TIME_CAR_CL',0.0,-10,10,0)
BETA_TIME_PT_REF = Beta('BETA_TIME_PT_REF',0.0,-10000,0,0 )
BETA_TIME_PT_CL = Beta('BETA_TIME_PT_CL',0.0,-10,10,0)
BETA_WAITING_TIME = Beta('BETA_WAITING_TIME',0.0,-10000,10000,0 )

TimePT_scaled  = DefineVariable('TimePT_scaled', TimePT   /  200 ,database)
TimeCar_scaled  = DefineVariable('TimeCar_scaled', TimeCar   /  200 ,database)
MarginalCostPT_scaled  = \
 DefineVariable('MarginalCostPT_scaled', MarginalCostPT   /  10 ,database)
CostCarCHF_scaled  = \
 DefineVariable('CostCarCHF_scaled', CostCarCHF   /  10 ,database)
distance_km_scaled  = \
 DefineVariable('distance_km_scaled', distance_km   /  5 ,database)
PurpHWH = DefineVariable('PurpHWH', TripPurpose == 1,database)
PurpOther = DefineVariable('PurpOther', TripPurpose != 1,database)

### Definition of utility functions:

BETA_TIME_PT = BETA_TIME_PT_REF * exp(BETA_TIME_PT_CL * CARLOVERS)

V0 = ASC_PT + \
     BETA_TIME_PT * TimePT_scaled + \
     BETA_WAITING_TIME * WaitingTimePT + \
     BETA_COST_HWH * MarginalCostPT_scaled * PurpHWH  +\
     BETA_COST_OTHER * MarginalCostPT_scaled * PurpOther

BETA_TIME_CAR = BETA_TIME_CAR_REF * exp(BETA_TIME_CAR_CL * CARLOVERS)

V1 = ASC_CAR + \
      BETA_TIME_CAR * TimeCar_scaled + \
      BETA_COST_HWH * CostCarCHF_scaled * PurpHWH  + \
      BETA_COST_OTHER * CostCarCHF_scaled * PurpOther 

V2 = ASC_SM + BETA_DIST * distance_km_scaled

# Associate utility functions with the numbering of alternatives
V = {0: V0,
     1: V1,
     2: V2}

# Associate the availability conditions with the alternatives.
# In this example all alternatives are available for each individual.
av = {0: 1,
      1: 1,
      2: 1}

# Conditional to omega, we have a logit model (called the kernel)
condprob = models.logit(V,av,Choice)

# We integrate over omega using numerical integration
loglike = log(Integrate(condprob * density,'omega'))

# Define level of verbosity
import biogeme.messaging as msg
logger = msg.bioMessage()
#logger.setSilent()
#logger.setWarning()
logger.setGeneral()
#logger.setDetailed()

# Create the Biogeme object
biogeme  = bio.BIOGEME(database,loglike)
biogeme.modelName = "04latentChoiceSeq"

# Estimate the parameters
results = biogeme.estimate(algorithm=None)

print(f"Estimated betas: {len(results.data.betaValues)}")
print(f"Final log likelihood: {results.data.logLike:.3f}")
print(f"Output file: {results.data.htmlFileName}")
results.writeLaTeX()
print(f"LaTeX file: {results.data.latexFileName}")






#######################################
#
# File: 02antithetic_halton.py
# Author: Michel Bierlaire, EPFL
# Date: Sat Jul 25 12:21:10 2015
#
#######################################
#

from biogeme import *
from headers import *

integrand = 0.5 * (exp(bioDraws('U')) + exp(1.0-bioDraws('U')))
simulatedI = MonteCarlo(integrand)

trueI = exp(1.0) - 1.0 

sampleVariance = \
  MonteCarlo(integrand*integrand) - simulatedI * simulatedI
stderr = (sampleVariance / 10000.0)**0.5
error = simulatedI - trueI

simulate = {'01_Simulated Integral': simulatedI,
            '02_Analytical Integral': trueI,
            '03_Sample variance': sampleVariance,
            '04_Std Error': stderr,
            '05_Error': error}

rowIterator('obsIter') 

BIOGEME_OBJECT.SIMULATE = Enumerate(simulate,'obsIter')

BIOGEME_OBJECT.PARAMETERS['NbrOfDraws'] = "10000"
BIOGEME_OBJECT.PARAMETERS['RandomDistribution'] = "HALTON"
__rowId__ = Variable('__rowId__')
BIOGEME_OBJECT.EXCLUDE = __rowId__ >= 1
BIOGEME_OBJECT.DRAWS = { 'U': 'UNIFORM'}



import unittest
import pandas as pd
import biogeme.expressions as ex
import biogeme.models as models
import biogeme.database as db

#import numpy as np

from pathlib import Path
import os

class testExpressions(unittest.TestCase):
    def setUp(self):
        df = pd.DataFrame({'Person':[1,1,1,2,2],
                           'Exclude':[0,0,1,0,1],
                           'Variable1':[10,20,30,40,50],
                           'Variable2':[100,200,300,400,500],
                           'Choice':[1,2,3,1,2],
                           'Av1':[0,1,1,1,1],
                           'Av2':[1,1,1,1,1],
                           'Av3':[0,1,1,1,1]})
        self.myData = db.Database('test',df)
        self.Person=ex.Variable('Person')
        self.Variable1=ex.Variable('Variable1')
        self.Variable2=ex.Variable('Variable2')
        self.Choice=ex.Variable('Choice')
        self.Av1=ex.Variable('Av1')
        self.Av2=ex.Variable('Av2')
        self.Av3=ex.Variable('Av3')
        self.beta1 = ex.Beta('beta1',1,None,None,0)
        self.beta2 = ex.Beta('beta2',2,None,None,0)
        self.beta3 = ex.Beta('beta3',3,None,None,1)
        self.beta4 = ex.Beta('beta4',2,None,None,1)        

    def test_DefineVariable(self):
        newvar = ex.DefineVariable('newvar',self.Variable1+self.Variable2,self.myData)
        cols = self.myData.data.columns
        added = 'newvar' in cols
        self.assertTrue(added)

    def test_expr1(self):
        expr1 = 2 * self.beta1 - ex.exp(-self.beta2) / (self.beta3 * (self.beta2 >= self.beta1))
        self.assertAlmostEqual(expr1.getValue(),1.954888238921129,5)
        res = expr1.getValue_c(self.myData)
        for v in res:
            self.assertAlmostEqual(v,1.954888238921129,5)

    def test_setOfBetas(self):
        expr1 = 2 * self.beta1 - ex.exp(-self.beta2) / (self.beta3 * (self.beta2 >= self.beta1))
        s = expr1.setOfBetas()
        self.assertSetEqual(s,{'beta1', 'beta2'})
        s = expr1.setOfBetas(free=False,fixed=True)        
        self.assertSetEqual(s,{'beta3'})
        s = expr1.setOfBetas(free=True,fixed=True)
        self.assertSetEqual(s,{'beta1', 'beta2', 'beta3'})

    def test_getElementaryExpression(self):
        expr1 = 2 * self.beta1 - ex.exp(-self.beta2) / (self.beta3 * (self.beta2 >= self.beta1))
        l = expr1.getElementaryExpression('beta2')
        self.assertEqual(l.name,'beta2')

    def test_expr2(self):
        expr2 = 2 * self.beta1 * self.Variable1 - ex.exp(-self.beta2*self.Variable2) / (self.beta3 * (self.beta2 >= self.beta1))
        res = expr2.getValue_c(self.myData)
        self.assertListEqual(res,[20.0, 40.0, 60.0, 80.0, 100.0])


    def test_dictOfBetas(self):
        expr2 = 2 * self.beta1 * self.Variable1 - ex.exp(-self.beta2*self.Variable2) / (self.beta3 * (self.beta2 >= self.beta1))
        b = expr2.dictOfBetas(free=True,fixed=True)
        # Note that the following checks only the labels. Its probably
        # good enough for our purpose.
        self.assertDictEqual(b,{'beta1': 0, 'beta2': self.beta2, 'beta3': self.beta3})

    def test_getClassName(self):
        expr2 = 2 * self.beta1 * self.Variable1 - ex.exp(-self.beta2*self.Variable2) / (self.beta3 * (self.beta2 >= self.beta1))
        c = expr2.getClassName()
        self.assertEqual(c,'Minus')

    def test_signature(self):
        expr2 = 2 * self.beta1 * self.Variable1 - ex.exp(-self.beta2*self.Variable2) / (self.beta3 * (self.beta2 >= self.beta1))
        expr2._prepareFormulaForEvaluation(self.myData)
        s = expr2.getSignature()
        self.assertEqual(len(s),17)

    def test_expr3(self):
        myDraws = ex.bioDraws('myDraws','UNIFORM')
        expr3 = ex.MonteCarlo(myDraws*myDraws)
        res = expr3.getValue_c(self.myData,numberOfDraws=100000)
        for v in res:
            self.assertAlmostEqual(v,1.0/3.0,2)


    def test_expr4(self):
        omega = ex.RandomVariable('omega')
        a = 0
        b = 1
        x = a + (b-a) / ( 1 + ex.exp(-omega))
        dx = (b-a) * ex.exp(-omega) * (1+ex.exp(-omega))**(-2) 
        integrand = x * x
        expr4 = ex.Integrate(integrand * dx /(b-a),'omega')
        res = expr4.getValue_c(self.myData)
        for v in res:
            self.assertAlmostEqual(v,1.0/3.0,2)

    def test_expr5(self):
        expr1 = 2 * self.beta1 - ex.exp(-self.beta2) / (self.beta3 * (self.beta2 >= self.beta1))
        expr2 = 2 * self.beta1 * self.Variable1 - ex.exp(-self.beta2*self.Variable2) / (self.beta3 * (self.beta2 >= self.beta1))
        expr5 = ex.Elem({1:expr1,2:expr2},self.Person) / 10
        res = expr5.getValue_c(self.myData)
        self.assertListEqual(res,[0.19548882389211292, 0.19548882389211292, 0.19548882389211292, 8.0, 10.0])

    def test_expr6(self):
        expr1 = 2 * self.beta1 - ex.exp(-self.beta2) / (self.beta3 * (self.beta2 >= self.beta1))
        expr2 = 2 * self.beta1 * self.Variable1 - ex.exp(-self.beta2*self.Variable2) / (self.beta3 * (self.beta2 >= self.beta1))
        omega = ex.RandomVariable('omega')
        a = 0
        b = 1
        x = a + (b-a) / ( 1 + ex.exp(-omega))
        dx = (b-a) * ex.exp(-omega) * (1+ex.exp(-omega))**(-2) 
        integrand = x * x
        expr4 = ex.Integrate(integrand * dx /(b-a),'omega')
        expr6 = ex.bioMultSum([expr1,expr2,expr4])
        res = expr6.getValue_c(self.myData)
        self.assertListEqual(res,[22.28822055098741, 42.28822055098741, 62.28822055098741, 82.2882205509874, 102.2882205509874])

    def test_expr7(self):
        V = {0:-self.beta1,1:-self.beta2,2:-self.beta1}
        av = {0:1,1:1,2:1}
        expr7 = ex.LogLogit(V,av,1)
        r = expr7.getValue()
        self.assertAlmostEqual(r,-1.861994804058251,5)
        expr8 = models.loglogit(V,av,1)
        res = expr8.getValue_c(self.myData)
        for v in res: 
            self.assertAlmostEqual(v,-1.861994804058251,5)

    def test_expr9(self):
        V = {0:-self.beta1,1:-self.beta2,2:-self.beta1}
        av = {0:1,1:1,2:1}
        expr8 = models.loglogit(V,av,1)
        expr9 = ex.Derive(expr8,'beta2')
        res = expr9.getValue_c(self.myData)
        for v in res:
            self.assertAlmostEqual(v,-0.8446375965030364,5)

    def test_expr10(self):
        expr10 = ex.bioNormalCdf(self.Variable1/10-1)
        res = expr10.getValue_c(self.myData)
        for i,j in zip(res,[0.5, 0.8413447460685283, 0.9772498680518218, 0.99865010196837, 0.9999683287581669]):
            self.assertAlmostEqual(i,j,5)

    def test_expr11(self):
        expr1 = 2 * self.beta1 - ex.exp(-self.beta2) / (self.beta3 * (self.beta2 >= self.beta1))
        expr2 = 2 * self.beta1 * self.Variable1 - ex.exp(-self.beta2*self.Variable2) / (self.beta3 * (self.beta2 >= self.beta1))
        expr5 = ex.Elem({1:expr1,2:expr2},self.Person) / 10
        expr10 = ex.bioNormalCdf(self.Variable1/10-1)
        expr11 = ex.bioMin(expr5,expr10)
        res = expr11.getValue_c(self.myData)
        for i,j in zip(res,[0.19548882389211292, 0.19548882389211292, 0.19548882389211292, 0.99865010196837, 0.9999683287581669]):
            self.assertAlmostEqual(i,j,5)

    def test_expr12(self):
        expr1 = 2 * self.beta1 - ex.exp(-self.beta2) / (self.beta3 * (self.beta2 >= self.beta1))
        expr2 = 2 * self.beta1 * self.Variable1 - ex.exp(-self.beta2*self.Variable2) / (self.beta3 * (self.beta2 >= self.beta1))
        expr5 = ex.Elem({1:expr1,2:expr2},self.Person) / 10
        expr10 = ex.bioNormalCdf(self.Variable1/10-1)
        expr12 = ex.bioMax(expr5,expr10)
        res = expr12.getValue_c(self.myData)
        for i,j in zip(res,[0.5, 0.8413447460685283, 0.9772498680518218, 8.0, 10.0]):
            self.assertAlmostEqual(i,j,5)

            
if __name__ == '__main__':
    unittest.main()



"""
Interface for the optimization algorithms.

:author: Michel Bierlaire
:date: Fri Aug 16 11:31:43 2019

"""



import numpy as np
import biogeme.exceptions as excep
import scipy.optimize as sc
import scipy.linalg as la
import biogeme.messaging as msg

from abc import ABC, abstractmethod

logger = msg.bioMessage()


class functionToMinimize:
    """This is an abstract class. The actual function to minimize must be
       implemented in a concrete class deriving from this one.

    """
    
    @abstractmethod
    def setVariables(self,x):

        """ Set the values of the variables for which the function has to b calculated. 
        
        :param x: values
        :type x: numpy.array
        """
        pass
    
    @abstractmethod
    def f(self, batch = None):
        """Calculate the value of the function
        
        :param batch: for data driven functions (such as a log
                      likelikood function), it is possible to
                      approximate the value of the function using a
                      sample of the data called a batch. This argument
                      is a value between 0 and 1 representing the
                      percentage of the data that should be used for
                      thre random batch. If None, the full data set is
                      used. Default: None pass
        :type batch: float
        
        :return: value of the function
        :rtype: float
        """
        pass
    
    @abstractmethod
    def f_g(self, batch = None):
        """Calculate the value of the function and the gradient
        
        :param batch: for data driven functions (such as a log
                      likelikood function), it is possible to
                      approximate the value of the function using a
                      sample of the data called a batch. This argument
                      is a value between 0 and 1 representing the
                      percentage of the data that should be used for
                      the random batch. If None, the full data set is
                      used. Default: None pass
        :type batch: float

        :return: value of the function and the gradient
        :rtype: tuple float, numpy.array
        """
        pass
    
    @abstractmethod
    def f_g_h(self, batch= None):
        """Calculate the value of the function, the gradient and the Hessian
        
        :param batch: for data driven functions (such as a log
                      likelikood function), it is possible to
                      approximate the value of the function using a
                      sample of the data called a batch. This argument
                      is a value between 0 and 1 representing the
                      percentage of the data that should be used for
                      the random batch. If None, the full data set is
                      used. Default: None pass
        :type batch: float
        
        :return: value of the function, the gradient and the Hessian
        :rtype: tuple float, numpy.array, numpy.array
        """
        pass
    
    @abstractmethod
    def f_g_bhhh(self, batch = None):
        """Calculate the value of the function, the gradient and the BHHH matrix
        
        :param batch: for data driven functions (such as a log
                      likelikood function), it is possible to
                      approximate the value of the function using a
                      sample of the data called a batch. This argument
                      is a value between 0 and 1 representing the
                      percentage of the data that should be used for
                      the random batch. If None, the full data set is
                      used. Default: None pass
        :type batch: float
        
        :return: value of the function, the gradient and the BHHH
        :rtype: tuple float, numpy.array, numpy.array
        """
        pass
    
def scipy(fct,initBetas,fixedBetas,betaIds,bounds,parameters=None):

    """Optimization interface for Biogeme, based on the scipy minimize function.

    :param fct: object to calculate the objective function and its derivatives.
    :type fct: optimization.functionToMinimize

    :param initBetas: initial value of the beta parameters
    :type initBetas: numpy.array
    :param fixedBetas: betas that stay fixed suring the optimization
    :type fixedBetas: numpy.array
    :param betaIds: internal identifiers of the non fixed betas
    :type betaIds: numpy.array
    :param bounds: list of tuples (ell,u) containing the lower and upper bounds
          for each free parameter
    :type bounds: list(tuple)
    :param parameters: dict of parameters to be transmitted to the 
         optimization routine. See the `scipy`_ documentation. 
 
    .. _`scipy`: https://docs.scipy.org/doc/scipy/reference/optimize.html

    :type parameters: dict(string:float or string)

    :return: tuple x, nit, nfev,message, where x is the solution found,
              nit is the number of iterations performed, nfev is the
              number of time that the objective function has been
              evaluated, and message is the diagnostic provided by the
              algorithm.
    :rtype: numpay.array, int, int, string

    """
    def f_and_grad(x):
        fct.setVariables(x)
        f, g = fct.f_g()
        return f, g

    # Absolute tolerance
    absgtol = 1.0e-7
    opts = {'ftol' : np.finfo(np.float64).eps, 'gtol': absgtol}
    if parameters is not None:
        opts = {**opts,**parameters}

    logger.general(f"Minimize with tol {opts['gtol']}")
    results =  sc.minimize(f_and_grad,initBetas,bounds=bounds,jac=True,options=opts)
    return results.x, results.nit,results.nfev,results.message


def schnabelEskow(A,tau=np.finfo(np.float64).eps**0.3333,taubar=np.finfo(np.float64).eps**0.6666,mu=0.1):

    """Modified Cholesky factorization by `Schnabel and Eskow (1999)`_.

    .. _`Schnabel and Eskow (1999)`: https://doi.org/10.1137/s105262349833266x

    If the matrix is 'safely' positive definite, the output is the
    classical Cholesky factor. If not, the diagonal elements are
    inflated in order to make it positive definite. The factor :math:`L`
    is such that :math:`A + E = PLL^TP^T`, where :math:`E` is a diagonal
    matrix contaninig the terms added to the diagonal, :math:`P` is a
    permutation matrix, and :math:`L` is w lower triangular matrix.

    :param A: matrix to factorize. Must be square and symmetric.
    :type A: numpy.array
    :param tau: tolerance factor. Default: :math:`\\varepsilon^{\\frac{1}{3}}`. See `Schnabel and Eskow (1999)`_
    :type tau: float
    :param taubar: tolerance factor. Default: :math:`\\varepsilon^{\\frac{2}{3}}`. See `Schnabel and Eskow (1999)`_
    :type taubar: float
    :param mu: tolerance factor. Default: 0.1.  See `Schnabel and Eskow (1999)`_
    :type mu: float

    :return: tuple :math:`L`, :math:`E`, :math:`P`, where :math:`A + E = PLL^TP^T`.
    :rtype: numpy.array, numpy.array, numpy.array

    :raises biogeme.exceptions.biogemeError: if the matrix A is not square.
    :raises biogeme.exceptions.biogemeError: if the matrix A is not symmetric.
    """
    def pivot(j):
        A[j, j] = np.sqrt(A[j, j])
        for i in range(j+1, dim):
            A[j, i] = A[i, j] = A[i, j] / A[j, j]
            A[i, j+1:i+1] -= A[i, j]*A[j+1:i+1, j]
            A[j+1:i+1, i] = A[i, j+1:i+1] 

    def permute(i,j):
        A[[i,j]] = A[[j,i]]
        E[[i,j]] = E[[j,i]]
        A[:,[i,j]] = A[:,[j,i]]
        P[:,[i,j]] = P[:,[j,i]]
        
    A = A.astype(np.float64)    
    dim = A.shape[0]
    if A.shape[1] != dim:
        raise excep.biogemeError("The matrix must be square")

    if not np.all(np.abs(A-A.T) < np.sqrt(np.finfo(np.float64).eps)):
        raise excep.biogemeError("The matrix must be symmetric")
    
    E = np.zeros(dim,dtype=np.float64)
    P = I = np.identity(dim)
    phaseOne = True
    gamma = abs(A.diagonal()).max()
    j = 0
    while j < dim and phaseOne is True:
        a_max = A.diagonal()[j:].max()
        a_min = A.diagonal()[j:].min()
        if (a_max < taubar*gamma or a_min < -mu*a_max):
            phaseOne = False
            break
        else:
            # Pivot on maximum diagonal of remaining submatrix
            i = j + np.argmax(A.diagonal()[j:])
            if i != j:
                # Switch rows and columns of i and j of A
                permute(i,j)
            if j < dim-1 and ((A.diagonal()[j+1:] - A[j+1:,j]**2/A.diagonal()[j]).min() < -mu*gamma):
                phaseOne = False # go to phase two
            else:
                # perform jth iteration of factorization
                pivot(j)
                j += 1

    # Phase two, A not positive-definite
    if not phaseOne:
        if j == dim - 1:
            E[-1] = delta = -A[-1,-1] + max(tau*(-A[-1,-1])/(1-tau), taubar*gamma)
            A[-1,-1] += delta
            A[-1,-1] = np.sqrt(A[-1,-1])
        else:
            deltaPrev = 0.0
            g = np.zeros(dim)
            k = j - 1  # k = number of iterations performed in phase one;
            # Calculate lower Gerschgorin bounds of A[k+1]
            for i in range(k + 1, dim):
                g[i] = A[i,i] - abs(A[i,k+1:i]).sum() - abs(A[i+1:dim,i]).sum()
            # Modified Cholesky Decomposition
            for j in range(k + 1, dim - 2):
                # Pivot on maximum lower Gerschgorin bound estimate
                i = j + np.argmax(g[j:])
                if i != j:
                    # Switch rows and columns of i and j of A
                    permute(i,j)
                # Calculate E[j,j] and add to diagonal
                norm_j = abs(A[j+1:dim,j]).sum()
                E[j] = delta = max(0,
                                   -A[j,j] + max(norm_j, taubar*gamma),
                                   deltaPrev)
                if delta > 0:
                    A[j,j] += delta
                    deltaPrev = delta         # deltaPrev will contain E_inf
                # Update Gerschgorin bound estimates
                if A[j,j] != norm_j:
                    temp = 1.0 - norm_j/A[j,j]
                    g[j+1:] += abs(A[j+1:,j])*temp
                # perform jth iteration of factorization
                pivot(j)

            # Final 2 by 2 submatrix
            e = np.linalg.eigvalsh(A[-2:,-2:])
            e.sort()
            E[-2] = E[-1] = delta = max(0,
                        -e[0] + max(tau*(e[1] - e[0])/(1 - tau),
                                    taubar*gamma),
                        deltaPrev)
            if delta > 0:
                A[-2,-2] += delta
                A[-1,-1] += delta
                deltaPrev = delta
            A[-2,-2] = np.sqrt(A[-2,-2])    # overwrites A[-2,-2]
            A[-1,-2] = A[-1,-2]/A[-2,-2]   # overwrites A[-1,-2]
            A[-2,-1] = A[-1,-2]
            A[-1,-1] = np.sqrt(A[-1,-1] - A[-1,-2]*A[-1,-2]) # overwrites A[-1,-1]
            

    return np.tril(A), np.diag(P @ E), P


def lineSearch(fct,x,f,g,d,alpha0 = 1.0,beta1 = 1.0e-4,beta2 = 0.99,lbd = 2.0):
    """
    Calculate a step along a direction that satisfies both Wolfe conditions

    
    :param fct: object to calculate the objective function and its derivatives.
    :type fct: optimization.functionToMinimize

    :param x: current iterate.
    :type x: numpy.array

    :param d: descent direction.
    :type d: numpy.array

    :param alpha0: first step to test.
    :type alpha0: float

    :param beta1: parameter of the first Wolfe condition.
    :type beta1: float

    :param beta2: parameter of the second Wolfe condition.
    :type beta2: float

    :param lbd: expansion factor for a short step.
    :type lbd: float

    :return: a step verifing both Wolfe conditions
    :rtype: float

    :raises biogeme.exceptions.biogemeError: if lbd :math:`\\leq` 1
    :raises biogeme.exceptions.biogemeError: if alpha0 :math:`\\leq` 0
    :raises biogeme.exceptions.biogemeError: if beta1 :math:`\\geq` beta2
    :raises biogeme.exceptions.biogemeError: if d is not a descent direction
    
    """
    if  lbd <= 1:
        raise excep.biogemeError(f"lambda is {lbd} and must be > 1")
    if  alpha0 <= 0:
        raise excep.biogemeError(f"alpha0 is {alpha0} and must be > 0")
    if  beta1 >= beta2:
        raise excep.biogemeError(f"Incompatible Wolfe cond. parsmeters: beta1={beta1} is greater than beta2={beta2}")

    nfev = 1
    deriv = np.inner(g,d)

    if deriv >= 0:
        raise excep.biogemeError(f"d is not a descent direction: {deriv} >= 0")
    i = 0
    alpha = alpha0
    alphal = 0
    alphar = np.finfo(np.float128).max
    finished = False
    while not finished:
        xnew = x + alpha * d ;
        fct.setVariables(xnew)
        fnew, gnew = fct.f_g()
        nfev += 1
        finished = True
        # First Wolfe condition violated?
        if fnew > f + alpha * beta1 * deriv:
            alphar = alpha ;
            alpha = (alphal + alphar) / 2.0 ;
            finished = False
        elif np.inner(gnew,d) < beta2 * deriv:
            alphal = alpha ;
            if alphar == np.finfo(np.float128).max:
                alpha = lbd * alpha ;
            else:
                alpha = (alphal + alphar) / 2.0
            finished = False
    return alpha,nfev

def relativeGradient(x,f,g,typx,typf):
    """ Calculates the relative gradients. 

    It is typically used for stopping criteria. 

    :param x: current iterate.
    :type x: numpy.array
    :param f: value of f(x)
    :type f: float
    :param g: :math:`\\nabla f(x)`, gradient of f at x 
    :type g: numpy.array
    :param typx: typical value for x.
    :type typx: numpy.array
    :param typf: typical value for f.
    :type typf: float

    :return: relative gradient

    .. math:: \\max_{i=1,\\ldots,n}\\frac{(\\nabla f(x))_i \\max(x_i,\\text{typx}_i)}{\\max(|f(x)|, \\text{typf})}

    :rtype: float
    """
    relgrad = np.array([g[i] * max(abs(x[i]),typx[i]) / max(abs(f),typf) for i in range(len(x))])
    result = abs(relgrad).max()
    if np.isfinite(result):
        return result
    else:
        return np.finfo(float).max
                           


def newtonLineSearch(fct,x0,eps=np.finfo(np.float64).eps**0.3333,maxiter=100):
    """
    Newton method with inexact line search (Wolfe conditions)

    :param fct: object to calculate the objective function and its derivatives.
    :type fct: optimization.functionToMinimize

    :param x0: starting point
    :type x0: numpy.array

    :param eps: the algorithm stops when this precision is reached. Default: :math:`\\varepsilon^{\\frac{1}{3}}`
    :type eps: float

    :param maxiter: the algorithm stops if this number of iterations is reached. Defaut: 100
    :type maxiter: int

    :return: x, nit, nf, diagnostic where
        
        - x is the solution generated by the algorithm,
        - nit is the number of iterations,
        - nf is the number of function evaluations,
        - diagnostic is the diagnostic provided by the algorithm.

    :rtype: numpy.array, int, int, string
    """

    xk = x0
    fct.setVariables(xk)
    f,g,H = fct.f_g_h()
    typx = np.ones(np.asarray(xk).shape)
    typf = max(np.abs(f) , 1.0)
    relgrad = relativeGradient(xk,f,g,typx,typf)
    if relgrad <= eps:
        message = f"Relative gradient = {relgrad} <= {eps}"
        return xk,0,1,message
        
    k = 0
    nfev = 0
    cont = True
    while cont:
        L,E,P = schnabelEskow(H)
        y3 = -P.T @ g
        y2 = la.solve_triangular(L,y3,lower=True)
        y1 = la.solve_triangular(L.T,y2,lower=False)
        d = P @ y1
        alpha,nfls = lineSearch(fct,xk,f,g,d)
        nfev += nfls
        xk = xk + alpha * d
        fct.setVariables(xk)
        f,g,H = fct.f_g_h()
        nfev += 1
        k += 1
        relgrad = relativeGradient(xk,f,g,typx,typf)
        if relgrad <= eps:
            message = f"Relative gradient = {relgrad} <= {eps}"
            cont = False
        if k == maxiter:
            message = f"Maximum number of iterations reached: {maxiter}"
            cont = False
        logger.detailed(f'{k} f={f:10.7g} relgrad={relgrad:6.2g} alpha={alpha:6.2g}')
    return xk,k,nfev,message


def newtonLineSearchForBiogeme(fct,initBetas,fixedBetas,betaIds,bounds,parameters=None):
    """Optimization interface for Biogeme, based on Newton method.

    :param fct: object to calculate the objective function and its derivatives.
    :type fct: optimization.functionToMinimize

    :param initBetas: initial value of the parameters.
    :type initBetas: numpy.array

    :param fixedBetas: betas that stay fixed suring the optimization.
    :type fixedBetas: numpy.array

    :param betaIds: internal identifiers of the non fixed betas.
    :type betaIds: numpy.array

    :param bounds: list of tuples (ell,u) containing the lower and upper bounds for each free parameter. Note that this algorithm does not support bound constraints. Therefore, all the bounds must be None.
    :type bounds: list(tuples)

    :param parameters: dict of parameters to be transmitted to the  optimization routine:             
         - tolerance: when the relative gradient is below that threshold, the algorithm has reached convergence (default:  :math:`\\varepsilon^{\\frac{1}{3}}`);
         - maxiter: the maximum number of iterations (default: 100).
    :type parameters: dict(string:float or int)

    :return: tuple x, nit, nfev, message, where 

            - x is the solution found,
            - nit is the number of iterations performed, 
            - nfev is the number of time that the objective function has been evaluated, and 
            - message is the diagnostic provided by the algorithm.
    :rtype: numpy.array, int, int, string

    :raises biogeme.exceptions.biogemeError: if bounds are imposed on the variables.

    """
    for l,u in bounds:
        if l is not None or u is not None:
            raise excep.biogemeError("This algorithm does not handle bound constraints. Remove the bounds, or select another algorithm.")

    tol = np.finfo(np.float64).eps**0.3333
    maxiter = 100
    if parameters is not None:
        if 'tolerance' in parameters:
            tol = parameters['tolerance']
        if 'maxiter' in parameters:
            maxiter = parameters['maxiter']

    logger.detailed("** Optimization: Newton with linesearch")
    return newtonLineSearch(fct,initBetas,eps=tol,maxiter=maxiter)

def trustRegionIntersection(dc,d,delta):
    """Calculates the intersection with the boundary of the trust region.
    
    Consider a trust region of radius :math:`\\delta`, centered at
    :math:`\\hat{x}`. Let :math:`x_c` be in the trust region, and
    :math:`d_c = x_c - \\hat{x}`, so that :math:`\\|d_c\\| \\leq
    \\delta`. Let :math:`x_d` be out of the trust region, and
    :math:`d_d = x_d - \\hat{x}`, so that :math:`\\|d_d\\| \\geq
    \\delta`.  We calculate :math:`\\lambda` such that

    .. math:: \\| d_c + \\lambda (d_d - d_c)\\| = \\delta

    :param dc: xc-xhat.
    :type dc: numpy.array
    :param d: dd - dc.
    :type d: numpy.array
    :param delta: radius of the trust region.
    :type delta: float

    :return: :math:`\\lambda` such that :math:`\\| d_c + \\lambda (d_d - d_c)\\| = \\delta`
    
    :rtype: float

    """
    a = np.inner(d,d)
    b = 2 * np.inner(dc,d)
    c = np.inner(dc,dc) - delta ** 2
    discriminant = b * b - 4.0 * a * c
    return (- b + np.sqrt(discriminant) ) / (2 * a)

def cauchyNewtonDogleg(g,H):
    """Calculate the Cauchy, the Newton and the dogleg points.

    The Cauchy point is defined as

    .. math:: d_c = - \\frac{\\nabla f(x)^T \\nabla f(x)}{\\nabla f(x)^T \\nabla^2 f(x)\\nabla f(x)} \\nabla f(x)

    The Newton point :math:`d_n` verifies Newton equation:

    .. math:: H_s d_n = - \\nabla f(x)

    where :math:`H_s` is a positive definite matrix generated with the method by `Schnabel and Eskow (1999)`_.

    The Dogleg point is 

    .. math:: d_d = \\eta d_n

    where

    .. math:: \\eta = 0.2 + 0.8 \\frac{\\alpha^2}{\\beta |\\nabla f(x)^T d_n|}

    and :math:`\\alpha= \\nabla f(x)^T \\nabla f(x)`, :math:`\\beta=\\nabla f(x)^T \\nabla^2 f(x)\\nabla f(x)`

    :param g: gradient :math:`\\nabla f(x)`

    :type g: numpy.array

    :param H: hessian :math:`\\nabla^2 f(x)`

    :type H: numpy.array

    :return: tuple with Cauchy point, Newton point, Dogleg point
    :rtype: numpy.array, numpy.array, numpy.array
    
    :raises biogeme.exceptions.biogemeError: if the quadratic model is not convex.

    """
    alpha = np.inner(g,g)
    beta = np.inner(g, H @ g)
    dc = - (alpha / beta ) * g
    L,E,P = schnabelEskow(H)
    if np.any(E):
        raise excep.biogemeError("The dogleg method requires a convex optimization problem.")
        
    y3 = -P.T @ g
    y2 = la.solve_triangular(L,y3,lower=True)
    y1 = la.solve_triangular(L.T,y2,lower=False)
    dn = P @ y1
    eta = 0.2 + (0.8 * alpha * alpha / (beta * abs(np.inner(g,dn)))) 
    return dc,dn,eta*dn
    
def dogleg(g,H,delta):
    """
    Find an approximation of the trust region subproblem using the dogleg method

    :param g: gradient of the quadratic model.
    :type g: numpy.array
    :param H: hessian of the quadratic model.
    :type H: numpy.array
    :param delta: radius of the trust region.
    :type delta: float

    :return: d, diagnostic where

          - d is an approximate solution of the trust region subproblem
          - diagnostic is the nature of the solution:

             * -2 if negative curvature along Newton direction
             * -1 if negative curvature along Cauchy direction (i.e. along the gradient)
             * 1 if partial Cauchy step 
             * 2 if Newton step
             * 3 if partial Newton step
             * 4 if Dogleg

    :rtype: numpy.array, int
    """

    dc,dn,dl = cauchyNewtonDogleg(g,H)

    
    # Check if the model is convex along the gradient direction

    alpha = np.inner(g,g)
    beta = np.inner(g,H @ g)
    if beta <= 0:
        dstar = -delta * g / np.sqrt(alpha)
        return dstar,-1

    # Compute the Cauchy point
    
    normdc = alpha * np.sqrt(alpha) / beta ;
    if normdc >= delta:
        # The Cauchy point is outside the trust
        # region. We move along the Cauchy
        # direction until the border of the trust
        # region.

        dstar = (delta / normdc) * dc
        return dstar, 1

    # Compute Newton point

    
    normdn = la.norm(dn)

    # Check the convexity of the model along Newton direction

    if np.inner(dn,H @ dn) <= 0.0:
        # Return the Cauchy point
        return dc,-2


    if normdn <= delta: 
        # Newton point is inside the trust region
        return dn,2


    # Compute the dogleg point

    eta = 0.2 + (0.8 * alpha * alpha / (beta * abs(np.inner(g,dn))))  

    partieldn = eta * la.norm(dn)
    
    if partieldn <= delta:
        # Dogleg point is inside the trust region
        dstar = (delta / normdn) * dn ;
        return dstar,3
  
    # Between Cauchy and dogleg
    nu = dl - dc
    lbd = trustRegionIntersection(dc,nu,delta)  
    dstar = dc + lbd * nu ;
    return dstar,4

def truncatedConjugateGradient(g,H,delta):
    """
    Find an approximation of the trust region subproblem using the 
    truncated conjugate gradient method

    :param g: gradient of the quadratic model.
    :type g: numpy.array
    :param H: hessian of the quadrartic model.
    :type H: numpy.array
    :param delta: radius of the trust region.
    :type delta: float

    :return: d, diagnostic, where 

          - d is the approximate solution of the trust region subproblem,
          - diagnostic is the nature of the solution:

            * 1 for convergence,
            * 2 if out of the trust region,
            * 3 if negative curvature detected.
            * 4 if a numerical problem has been encountered

    :rtype: numpy.array, int
    """
    tol = 1.0e-6
    n = len(g)
    xk = np.zeros(n)
    gk = g 
    dk = -gk
    for k in range(n):
        try:
            curv = np.inner(dk,H @ dk)
            if  curv <= 0:
                # Negative curvature has been detected
                type = 3
                a = np.inner(dk,dk) 
                b = 2 * np.inner(xk,dk)
                c = np.inner(xk,xk) - delta * delta
                rho = b * b - 4 * a * c ;
                step = xk + ((-b + np.sqrt(rho)) / (2*a)) * dk
                return step,type
            alphak = - np.inner(dk,gk) / curv
            xkp1 = xk + alphak * dk
            if np.isnan(xkp1).any() or la.norm(xkp1) > delta:
                # Out of the trust region
                type = 2 
                a = np.inner(dk,dk) 
                b = 2 * np.inner(xk,dk)
                c = np.inner(xk,xk) - delta * delta
                rho = b * b - 4 * a * c ;
                step = xk + ((-b + np.sqrt(rho)) / (2*a)) * dk
                return step, type
            xk = xkp1 ;
            gkp1 = H @ xk + g
            betak = np.inner(gkp1,gkp1) / np.inner(gk,gk)
            dk = -gkp1 + betak * dk
            gk = gkp1
            if la.norm(gkp1) <= tol:
              type = 1
              step = xk
              return step,type
        except:
            # Numerical problem
            type = 4
            a = np.inner(dk,dk) 
            b = 2 * np.inner(xk,dk)
            c = np.inner(xk,xk) - delta * delta
            rho = b * b - 4 * a * c ;
            step = xk + ((-b + np.sqrt(rho)) / (2*a)) * dk
            return step,type
    type = 1
    step = xk
    return step, type


def newtonTrustRegion(fct,x0,delta0=1.0,eps=np.finfo(np.float64).eps**0.3333,dl=False,maxiter=1000,eta1 = 0.01,eta2 = 0.9):
    """Newton method with trust region

    :param fct: object to calculate the objective function and its derivatives.
    :type fct: optimization.functionToMinimize

    :param x0: starting point
    :type x0: numpy.array

    :param delta0: initial radius of the trust region. Default: 100.
    :type delta0: float

    :param eps: the algorithm stops when this precision is reached. Default: :math:`\\varepsilon^{\\frac{1}{3}}`
    :type eps: float

    :param dl: If True, the Dogleg method is used to solve the
       trut region subproblem. If False, the truncated conjugate
       gradient is used. Default: False.
    :type dl: bool

    :param maxiter: the algorithm stops if this number of iterations is reached. Default: 1000.
    :type maxiter: int

    :param eta1: threshold for failed iterations. Default: 0.01.
    :type eta1: float

    :param eta2: threshold for very successful iterations. Default 0.9.
    :type eta2: float

    :return: tuple x, nit, nfev, message, where x is the solution found,
              nit is the number of iterations performed, nfev is the
              number of time that the objective function has been
              evaluated, and message is the diagnostic provided by the
              algorithm.
    :rtype: numpay.array, int, int, string

    """
    k = 0
    xk = x0
    fct.setVariables(xk)
    f,g,H = fct.f_g_h()
    typx = np.ones(np.asarray(xk).shape)
    typf = max(np.abs(f) , 1.0)
    relgrad = relativeGradient(xk,f,g,typx,typf)
    if relgrad <= eps:
        message = f"Relative gradient = {relgrad} <= {eps}"
        return xk,0,1,message
    delta = delta0
    nfev = 0
    cont = True
    maxDelta = np.finfo(float).max
    minDelta = np.finfo(float).eps
    rho = 0.0
    while cont:
        k += 1
        if dl:
            step,type = dogleg(g,H,delta)
        else: 
            step,type = truncatedConjugateGradient(g,H,delta)
        xc = xk + step
        fct.setVariables(xc)
        # Calculate the value of the function
        fc = fct.f()
        nfev += 1
        num = f - fc;
        denom = -np.inner(step,g) - 0.5 * np.inner(step,H @ step)
        rho = num / denom
        if rho < eta1:
            # Failure: reduce the trust region
            delta = la.norm(step) / 2.0
            status = "-"
        else:
            # Candidate accepted
            fc, gc, Hc = fct.f_g_h()
            xk = xc
            f = fc
            g = gc
            H = Hc
            if rho >= eta2:
                # Enlarge the trust region
                delta = min(2 * delta,maxDelta)
                status = "++"
            else:
                status = "+"
            relgrad = relativeGradient(xk,f,g,typx,typf)
            if relgrad <= eps:
                message = f"Relative gradient = {relgrad} <= {eps}"
                cont = False
        if delta <= minDelta:
            message = f"Trust region is too small: {delta}"
            cont = False
        if k == maxiter:
            message = f"Maximum number of iterations reached: {maxiter}"
            cont = False
        logger.detailed(f"{k} f={f:10.7g} relgrad={relgrad:6.2g} delta={delta:6.2g} rho={rho:6.2g} {status}")
    return xk,k,nfev,message

def newtonTrustRegionForBiogeme(fct,initBetas,fixedBetas,betaIds,bounds,parameters=None):
    """Optimization interface for Biogeme, based on Newton method with TR.

    :param fct: object to calculate the objective function and its derivatives.
    :type fct: optimization.functionToMinimize

    :param initBetas: initial value of the parameters.
    :type initBetas: numpy.array

    :param fixedBetas: betas that stay fixed suring the optimization.
    :type fixedBetas: numpy.array

    :param betaIds: internal identifiers of the non fixed betas.
    :type betaIds: numpy.array

    :param bounds: list of tuples (ell,u) containing the lower and upper bounds for each free parameter. Note that this algorithm does not support bound constraints. Therefore, all the bounds must be None.
    :type bounds: list(tuples)

    :param parameters: dict of parameters to be transmitted to the  optimization routine:             
         - tolerance: when the relative gradient is below that threshold, the algorithm has reached convergence (default:  :math:`\\varepsilon^{\\frac{1}{3}}`);
         - maxiter: the maximum number of iterations (default: 100).
         - dogleg: if True, the trust region subproblem is solved using 
                   the Dogleg method. If False, it is solved using the 
                   truncated conjugate gradient method (default: False).
         - radius: the initial radius of the truat region (default: 1.0).
    :type parameters: dict(string:float or int)

    :return: tuple x, nit, nfev, message, where 

            - x is the solution found,
            - nit is the number of iterations performed, 
            - nfev is the number of time that the objective function has been evaluated, and 
            - message is the diagnostic provided by the algorithm.
    :rtype: numpy.array, int, int, string

    :raises biogeme.exceptions.biogemeError: if bounds are imposed on the variables.

    """
    for l,u in bounds:
        if l is not None or u is not None:
            raise excep.biogemeError("This algorithm does not handle bound constraints. Remove the bounds, or select another algorithm.")

    tol = np.finfo(np.float64).eps**0.3333
    maxiter = 100
    dogleg = False
    radius = 1.0
    if parameters is not None:
        if 'tolerance' in parameters:
            tol = parameters['tolerance']
        if 'maxiter' in parameters:
            maxiter = parameters['maxiter']
        if 'dogleg' in parameters:
            dogleg = parameters['dogleg']
        if 'radius' in parameters:
            radius = parameters['radius']

    logger.detailed("** Optimization: Newton with trust region")
    return newtonTrustRegion(fct,x0=initBetas,delta0=radius,eps=tol,dl=dogleg,maxiter=maxiter)

def bfgs(H,d,y):
    """Update the BFGS matrix. Formula (13.12) of `Bierlaire (2015)`_ where the method proposed by `Powell (1977)`_ is applied

    .. _`Bierlaire (2015)`: http://optimizationprinciplesalgorithms.com/
    .. _`Powell (1977)`: https://link.springer.com/content/pdf/10.1007/BFb0067703.pdf

    :param H: current approximation of the inverse of the Hessian
    :type H: numpy.array (2D)

    :param d: difference between two consecutive iterates.
    :type d: numpy.array (1D)

    :param y: difference between two consecutive gradients.
    :type y: numpy.array (1D)

    :return: updated approximation of the inverse of the Hessian. 
    :rtype: numpy.array (2D)

    """
    n = len(d)

    Hd = H @ d
    dHd = np.inner(d, Hd)
    denom = np.inner(d,y)
    if denom >= 0.2 * dHd:
        eta = y
    else:
        theta = 0.8 * dHd / (dHd - denom)
        eta = theta * y + (1 - theta) * Hd

    return H - np.outer(Hd, Hd) / dHd + np.outer(eta, eta) / np.inner(d, eta)


def inverseBfgs(Hinv,d,y):
    """ Update the inverse BFGS matrix. Formula (13.13) of `Bierlaire (2015)`_

    .. _`Bierlaire (2015)`: http://optimizationprinciplesalgorithms.com/

    :param Hinv: current approximation of the inverse of the Hessian
    :type Hinv: numpy.array (2D)

    :param d: difference between two consecutive iterates.
    :type d: numpy.array (1D)

    :param y: difference between two consecutive gradients.
    :type y: numpy.array (1D)

    :return: updated approximation of the inverse of the Hessian. 
    :rtype: numpy.array (2D)
    """
    n = len(d)

    denom = np.inner(d,y)
    if denom <= 0.0:
        logger.warning(f"Unable to perform BFGS update as d'y = {denom} <= 0")
        return Hinv
    dy = np.outer(d, y)
    yd = np.outer(y, d)
    dd = np.outer(d, d)
    I = np.identity(n)
    return ((I - ( dy / denom )) @ Hinv @ (I - ( yd / denom ))) + dd / denom

def bfgsLineSearch(fct,x0,initBfgs=None,eps=np.finfo(np.float64).eps**0.3333,maxiter=1000):
    """BFGS method with inexact line search (Wolfe conditions)

    :param fct: object to calculate the objective function and its derivatives.
    :type fct: optimization.functionToMinimize

    :param x0: starting point
    :type x0: numpy.array

    :param initBfgs: matrix used to initialize BFGS. If None, the
                     identity matrix is used. Default: None.
    :type initBfgs: numpy.array

    :param eps: the algorithm stops when this precision is reached. Default: :math:`\\varepsilon^{\\frac{1}{3}}`
    :type eps: float

    :param maxiter: the algorithm stops if this number of iterations is reached. Default: 1000
    :type maxiter: int

    :return: x, nit, nf, diagnostic where
        
        - x is the solution generated by the algorithm,
        - nit is the number of iterations,
        - nf is the number of function evaluations,
        - diagnostic is the diagnostic provided by the algorithm.

    :rtype: numpy.array, int, int, string

    :raises biogeme.exceptions.biogemeError: if the dimensions of the matrix initBfgs do not match the length of x0.

    """

    n = len(x0)
    xk = x0
    fct.setVariables(xk)
    f,g = fct.f_g()
    if initBfgs is None:
        Hinv = np.identity(n)
    else:
        if initBfgs.shape != (n,n):
            raise excep.biogemeError(f'BFGS must be initialized with a {n}x{n} matrix and not a {initBfgs.shape[0]}x{initBfgs.shape[1]} matrix.')
        Hinv = initBfgs
    typx = np.ones(np.asarray(xk).shape)
    typf = max(np.abs(f) , 1.0)
    relgrad = relativeGradient(xk,f,g,typx,typf)
    if relgrad <= eps:
        message = f"Relative gradient = {relgrad} <= {eps}"
        return xk,0,1,message
    k = 0
    nfev = 0
    cont = True
    while cont:
        d = -Hinv @ g
        alpha,nfls = lineSearch(fct,xk,f,g,d)
        nfev += nfls
        delta = alpha * d
        xk = xk + delta
        gprev = g
        fct.setVariables(xk)
        f,g = fct.f_g()
        Hinv = inverseBfgs(Hinv,delta,g-gprev)
        nfev += 1
        k += 1
        relgrad = relativeGradient(xk,f,g,typx,typf)
        if relgrad <= eps:
            message = f"Relative gradient = {relgrad} <= {eps}"
            cont = False
        if k == maxiter:
            message = f"Maximum number of iterations reached: {maxiter}"
            cont = False
        logger.detailed(f'{k} f={f:10.7g} relgrad={relgrad:6.2g} alpha={alpha:6.2g}')
    return xk,k,nfev,message

def bfgsLineSearchForBiogeme(fct,initBetas,fixedBetas,betaIds,bounds,parameters=None):
    """Optimization interface for Biogeme, based on BFGS quasi-Newton method with LS.
                       
    :param fct: object to calculate the objective function and its derivatives.
    :type fct: optimization.functionToMinimize

    :param initBetas: initial value of the parameters.
    :type initBetas: numpy.array

    :param fixedBetas: betas that stay fixed suring the optimization.
    :type fixedBetas: numpy.array

    :param betaIds: internal identifiers of the non fixed betas.
    :type betaIds: numpy.array

    :param bounds: list of tuples (ell,u) containing the lower and upper bounds for each free parameter. Note that this algorithm does not support bound constraints. Therefore, all the bounds must be None.
    :type bounds: list(tuples)

    :param parameters: dict of parameters to be transmitted to the  optimization routine:  
           
         - tolerance: when the relative gradient is below that threshold, the algorithm has reached convergence (default:  :math:`\\varepsilon^{\\frac{1}{3}}`);
         - maxiter: the maximum number of iterations (default: 100).
         - initBfgs: the positive definite matrix that initalizes the BFGS updates. If None, the identity matrix is used. Default: None. 

    :type parameters: dict(string:float or int)

    :return: tuple x, nit, nfev, message, where 

            - x is the solution found,
            - nit is the number of iterations performed, 
            - nfev is the number of time that the objective function has been evaluated, and 
            - message is the diagnostic provided by the algorithm.
    :rtype: numpy.array, int, int, string

    :raises biogeme.exceptions.biogemeError: if bounds are imposed on the variables.

    """
    for l,u in bounds:
        if l is not None or u is not None:
            raise excep.biogemeError("This algorithm does not handle bound constraints. Remove the bounds, or select another algorithm.")

    tol = np.finfo(np.float64).eps**0.3333
    maxiter = 100
    initBfgs = None
    if parameters is not None:
        if 'tolerance' in parameters:
            tol = parameters['tolerance']
            if 'maxiter' in parameters:
                maxiter = parameters['maxiter']
            if 'initBfgs' in parameters:
                initBfgs = parameters['initBfgs']

    logger.detailed("** Optimization: BFGS with line search")
    return bfgsLineSearch(fct,x0=initBetas,initBfgs=initBfgs,eps=tol,maxiter=maxiter)


def bfgsTrustRegion(fct,x0,initBfgs=None,delta0=1.0,eps=np.finfo(np.float64).eps**0.3333,dl=False,maxiter=1000,eta1 = 0.01,eta2 = 0.9):
    """BFGS method with trust region

    :param fct: object to calculate the objective function and its derivatives.
    :type fct: optimization.functionToMinimize

    :param x0: starting point
    :type x0: numpy.array

    :param initBfgs: matrix used to initialize BFGS. If None, the
                     identity matrix is used. Default: None.
    :type initBfgs: numpy.array

    :param delta0: initial radius of the trust region. Default: 100.
    :type delta0: float

    :param eps: the algorithm stops when this precision is reached. Default: :math:`\\varepsilon^{\\frac{1}{3}}`
    :type eps: float

    :param dl: If True, the Dogleg method is used to solve the
       trut region subproblem. If False, the truncated conjugate
       gradient is used. Default: False.
    :type dl: bool

    :param maxiter: the algorithm stops if this number of iterations is reached. Default: 1000.
    :type maxiter: int

    :param eta1: threshold for failed iterations. Default: 0.01.
    :type eta1: float

    :param eta2: threshold for very successful iterations. Default 0.9.
    :type eta2: float

    :return: tuple x, nit, nfev, message, where x is the solution found,
              nit is the number of iterations performed, nfev is the
              number of time that the objective function has been
              evaluated, and message is the diagnostic provided by the
              algorithm.
    :rtype: numpay.array, int, int, string

    :raises biogeme.exceptions.biogemeError: if the dimensions of the matrix initBfgs do not match the length of x0.

    """
    k = 0
    xk = x0
    n = len(x0)
    fct.setVariables(xk)
    f,g = fct.f_g()
    if initBfgs is None:
        H = np.identity(n)
    else:
        if initBfgs.shape != (n,n):
            raise excep.biogemeError(f'BFGS must be initialized with a {n}x{n} matrix and not a {initBfgs.shape[0]}x{initBfgs.shape[1]} matrix.')
        H = initBfgs
    typx = np.ones(np.asarray(xk).shape)
    typf = max(np.abs(f) , 1.0)
    relgrad = relativeGradient(xk,f,g,typx,typf)
    if relgrad <= eps:
        message = f"Relative gradient = {relgrad} <= {eps}"
        return xk, 0, 1, message
    delta = delta0
    nfev = 1
    cont = True
    maxDelta = np.finfo(float).max
    minDelta = np.finfo(float).eps
    rho = 0.0
    while cont:
        k += 1
        if dl:
            step,type = dogleg(g,H,delta)
        else: 
            step,type = truncatedConjugateGradient(g,H,delta)
        xc = xk + step
        fct.setVariables(xc)
        # Calculate the value of the function
        fc = fct.f()
        nfev += 1
        if fc >= f:
            delta = la.norm(step) / 2.0
            status = "-"
        else:
            num = f - fc;
            denom = -np.inner(step,g) - 0.5 * np.inner(step,H @ step)
            rho = num / denom
            if rho < eta1:
                # Failure: reduce the trust region
                delta = la.norm(step) / 2.0
                status = "-"
            else:
                # Candidate accepted
                fc, gc = fct.f_g()
                d = xc - xk
                y = gc - g
                xk = xc
                f = fc
                g = gc
                H = bfgs(H, d, y)
                if rho >= eta2:
                    # Enlarge the trust region
                    delta = min(2 * delta,maxDelta)
                    status = "++"
                else:
                    status = "+"
                relgrad = relativeGradient(xk,f,g,typx,typf)
                if relgrad <= eps:
                    message = f"Relative gradient = {relgrad} <= {eps}"
                    cont = False
        if delta <= minDelta:
            message = f"Trust region is too small: {delta}"
            cont = False
        if k == maxiter:
            message = f"Maximum number of iterations reached: {maxiter}"
            cont = False
        logger.detailed(f"{k} f={f:10.7g} relgrad={relgrad:6.2g} delta={delta:6.2g} rho={rho:6.2g} {status}")
                
    return xk, k, nfev, message


def bfgsTrustRegionForBiogeme(fct,initBetas,fixedBetas,betaIds,bounds,parameters=None):
    """Optimization interface for Biogeme, based on Newton method with TR.

    :param fct: object to calculate the objective function and its derivatives.
    :type fct: optimization.functionToMinimize

    :param initBetas: initial value of the parameters.
    :type initBetas: numpy.array

    :param fixedBetas: betas that stay fixed suring the optimization.
    :type fixedBetas: numpy.array

    :param betaIds: internal identifiers of the non fixed betas.
    :type betaIds: numpy.array

    :param bounds: list of tuples (ell,u) containing the lower and upper bounds for each free parameter. Note that this algorithm does not support bound constraints. Therefore, all the bounds must be None.
    :type bounds: list(tuples)

    :param parameters: dict of parameters to be transmitted to the  optimization routine:             
         - tolerance: when the relative gradient is below that threshold, the algorithm has reached convergence (default:  :math:`\\varepsilon^{\\frac{1}{3}}`);
         - maxiter: the maximum number of iterations (default: 100).
         - dogleg: if True, the trust region subproblem is solved using 
                   the Dogleg method. If False, it is solved using the 
                   truncated conjugate gradient method (default: False).
         - radius: the initial radius of the truat region (default: 1.0).
         - initBfgs: the positive definite matrix that initalizes the BFGS updates. If None, the identity matrix is used. Default: None. 
 

    :type parameters: dict(string:float or int)

    :return: tuple x, nit, nfev, message, where 

            - x is the solution found,
            - nit is the number of iterations performed, 
            - nfev is the number of time that the objective function has been evaluated, and 
            - message is the diagnostic provided by the algorithm.
    :rtype: numpy.array, int, int, string

    :raises biogeme.exceptions.biogemeError: if bounds are imposed on the variables.

    """
    for l,u in bounds:
        if l is not None or u is not None:
            raise excep.biogemeError("This algorithm does not handle bound constraints. Remove the bounds, or select another algorithm.")

    tol = np.finfo(np.float64).eps**0.3333
    maxiter = 100
    dogleg = False
    radius = 1.0
    initBfgs = None
    if parameters is not None:
        if 'tolerance' in parameters:
            tol = parameters['tolerance']
        if 'maxiter' in parameters:
            maxiter = parameters['maxiter']
        if 'dogleg' in parameters:
            dogleg = parameters['dogleg']
        if 'radius' in parameters:
            radius = parameters['radius']
        if 'initBfgs' in parameters:
            initBfgs= parameters['initBfgs']
            

    logger.detailed("** Optimization: BFGS with trust region")
    return bfgsTrustRegion(fct,x0=initBetas,initBfgs=initBfgs,delta0=radius,eps=tol,dl=dogleg,maxiter=maxiter)




"""File 16panelDiscreteSocioEco.py

:author: Michel Bierlaire, EPFL
:date: Sun Sep  8 19:40:48 2019

 Example of a discrete mixture of logit models, also called latent class model.
 The class membership model includes socio-economic variables. 
 The datafile is organized as panel data.
 Three alternatives: Train, Car and Swissmetro
 SP data
"""
import pandas as pd
import biogeme.database as db
import biogeme.biogeme as bio
import biogeme.models as models
from biogeme.expressions import Beta, DefineVariable, bioDraws, PanelLikelihoodTrajectory, MonteCarlo, log

# Read the data
df = pd.read_csv("swissmetro.dat",sep='\t')
database = db.Database("swissmetro",df)

# They are organized as panel data. The variable ID identifies each individual.
database.panel("ID")

# The Pandas data structure is available as database.data. Use all the
# Pandas functions to invesigate the database
#print(database.data.describe())

# The following statement allows you to use the names of the variable
# as Python variable.
globals().update(database.variables)

# Removing some observations can be done directly using pandas.
#remove = (((database.data.PURPOSE != 1) & (database.data.PURPOSE != 3)) | (database.data.CHOICE == 0))
#database.data.drop(database.data[remove].index,inplace=True)

# Here we use the "biogeme" way for backward compatibility
exclude = (( PURPOSE != 1 ) * (  PURPOSE   !=  3  ) +  ( CHOICE == 0 )) > 0
database.remove(exclude)

# Parameters to be estimated
ASC_CAR = Beta('ASC_CAR',0.136,None,None,0)
ASC_TRAIN = Beta('ASC_TRAIN',-1,None,None,0)
ASC_SM = Beta('ASC_SM',0,None,None,1)
B_TIME = Beta('B_TIME',-6.3,None,0,0)
B_COST = Beta('B_COST',-3.29,None,0,0)
SIGMA_CAR = Beta('SIGMA_CAR',3.7,None,None,0)
SIGMA_SM = Beta('SIGMA_SM',0.759,None,None,0)
SIGMA_TRAIN = Beta('SIGMA_TRAIN',3.02,None,None,0)

# Define random parameters, normally distributed across individuals,
# designed to be used for Monte-Carlo simulation
EC_CAR = SIGMA_CAR * bioDraws('EC_CAR','NORMAL')
EC_SM = SIGMA_SM * bioDraws('EC_SM','NORMAL')
EC_TRAIN = SIGMA_TRAIN * bioDraws('EC_TRAIN','NORMAL')

# Definition of new variables
SM_COST =  SM_CO   * (  GA   ==  0  ) 
TRAIN_COST =  TRAIN_CO   * (  GA   ==  0  )

# Definition of new variables: adding columns to the database 
TRAIN_TT_SCALED = DefineVariable('TRAIN_TT_SCALED',\
                                 TRAIN_TT / 100.0,database)
TRAIN_COST_SCALED = DefineVariable('TRAIN_COST_SCALED',\
                                   TRAIN_COST / 100,database)
SM_TT_SCALED = DefineVariable('SM_TT_SCALED', SM_TT / 100.0,database)
SM_COST_SCALED = DefineVariable('SM_COST_SCALED', SM_COST / 100,database)
CAR_TT_SCALED = DefineVariable('CAR_TT_SCALED', CAR_TT / 100,database)
CAR_CO_SCALED = DefineVariable('CAR_CO_SCALED', CAR_CO / 100,database)

# Utility functions for latent class 1, where the time coefficient is zero
V11 = ASC_TRAIN + B_COST * TRAIN_COST_SCALED  + EC_TRAIN
V12 = ASC_SM + B_COST * SM_COST_SCALED + EC_SM
V13 = ASC_CAR + B_COST * CAR_CO_SCALED + EC_CAR
V1 = {1: V11,
      2: V12,
      3: V13}

# Utility functions for latent class 2, whete the time coefficient is estimated
V21 = ASC_TRAIN + B_TIME * TRAIN_TT_SCALED + B_COST * TRAIN_COST_SCALED + EC_TRAIN
V22 = ASC_SM + B_TIME * SM_TT_SCALED + B_COST * SM_COST_SCALED + EC_SM
V23 = ASC_CAR + B_TIME * CAR_TT_SCALED + B_COST * CAR_CO_SCALED + EC_CAR
V2 = {1: V21,
      2: V22,
      3: V23}


# Associate the availability conditions with the alternatives
CAR_AV_SP =  DefineVariable('CAR_AV_SP',CAR_AV  * (  SP   !=  0  ),database)
TRAIN_AV_SP =  DefineVariable('TRAIN_AV_SP',TRAIN_AV  * (  SP   !=  0  ),database)
av = {1: TRAIN_AV_SP,
      2: SM_AV,
      3: CAR_AV_SP}

# Class membership model
CLASS_CTE = Beta('CLASS_CTE',0,None,None,0)
CLASS_INC  = Beta('CLASS_INC',0,None,None,0)
W1 = CLASS_CTE + CLASS_INC * INCOME
probClass1 = models.logit({1:W1,2:0},None,1)
probClass2 = models.logit({1:W1,2:0},None,2)

# The choice model is a discrete mixture of logit, with availability conditions
# Conditional to the random variables, likelihood if the individual is
# in class 1
prob1 = PanelLikelihoodTrajectory(models.logit(V1,av,CHOICE))

# The choice model is a discrete mixture of logit, with availability conditions
# Conditional to the random variables, likelihood if the individual is
# in class 2
prob2 = PanelLikelihoodTrajectory(models.logit(V2,av,CHOICE))

# Conditional to the random variables, likelihood for the individual.
probIndiv = probClass1 * prob1 + probClass2 * prob2

# We integrate over the random variables using Monte-Carlo
logprob = log(MonteCarlo(probIndiv))

# Define level of verbosity
import biogeme.messaging as msg
logger = msg.bioMessage()
#logger.setSilent()
#logger.setWarning()
logger.setGeneral()
#logger.setDetailed()

# Create the Biogeme object
biogeme  = bio.BIOGEME(database,logprob)
biogeme.modelName = "16panelDiscreteSocioEco"

# Estimate the parameters. 
results = biogeme.estimate()
pandasResults = results.getEstimatedParameters()
print(pandasResults)





from biogeme.expressions import *

## \file 
# Implements various models.
# \author Michel Bierlaire
# \date Fri Mar 29 17:13:14 2019

def logit(V,av,i):
    """The logit model \f[ \frac{a_i e^{V_i}}{\sum_{i=1}^J a_j e^{V_j}} \f]

    Args:

       V: dict of objects of type biogeme.expressions.Expression  representing the
    utility functions of each alternative, indexed by numerical ids.

       av: dict of objects of type biogeme.expressions.Expression  representing
    the availability of each alternative (\f$a_i\f$ in the above formula), indexed by numerical ids. Must be consistent with V, or None. In this case, all alternatives are supposed to be always available.

       i: id of the alternative for which the probability must be calculated.

    Return:
       choice probability of alternative number i. 
    """

    if av is None:
        return exp(bioLogLogitFullChoiceSet(V,av=None,choice=i))
    else:
        return exp(bioLogLogit(V,av,i))


def boxcox(x,l):
    """
    Box-Cox transform: \f[ B(x,\ell) = \frac{x^{\ell}-1}{\ell} \f]. It has the property that \f[\lim_{\ell \to 0} B(x,\ell)=\log(x). \f]

    Args:
      x: a variable to transform.
      l: parameter of the transformation.
    
    Return: 
       the Box-Cox transform
    """
    return (x**l-1.0)/l

def piecewise(variable,thresholds):
    """ Generate the variables to include in a piecewise linear specification. If there are K thresholds, K+1 variables are generated. If \f$ t \f$ is the variable of interest, for each interval \f$[a:a+b[\f$ we define a variable defined as:  
\f[
x_{Ti} =
\left\{
  \begin{array}{ll}
 0 & \text{if } t < a \\
 t-a & \text{if } a \leq t < a+b \\
 b  & \text{otherwise}
  \end{array}
\right. \;\;\;x_{Ti} = \max(0,\min(t-a,b)) 
\f]

    Args:
       variable: variable for which we need the piecewise linear transform.
       thresholds: list of thresholds

    Return:
       list of variables to for the piecewise linear specification.
    """
    n = len(thresholds)
    results = [bioMin(variable,thresholds[0])]
    for i in range(0,n-1):
        b = thresholds[i+1]-thresholds[i]
        results += [bioMax(Numeric(0),bioMin(variable - thresholds[i],b))]
    results += [bioMax(0,variable - thresholds[-1])]
    return results
    

def logmev(V,Gi,av,choice) :
    """
     Log of the choice probability for a MEV model.
    
     Args: 
        V: A dictionary mapping each alternative id with the expression of the utility function.
        Gi: A dictionary mapping each alternative id with the function
        \f[
        \frac{\partial G}{\partial y_i}(e^{V_1},\ldots,e^{V_J})
        \f]
        where \f$G\f$ is the MEV generating function. If an alternative \f$i\f$ is not available, then \f$G_i = 0\f$.
      av: A dictionary mapping each alternative id with its availability condition. If None, all alternatives are considered available.
        choice: Object of type biogeme.expressions.Expression  producing the id of the chosen alternative.

    Return:
        Log of the choice probability of the MEV model, given by
  \f[
    V_i + \ln G_i(e^{V_1},\ldots,e^{V_J}) - \log\left(\sum_j e^{V_j + \ln G_j(e^{V_1},\ldots,e^{V_J})}\right)
  \f]
    """
    #    H = {}
    #    for i,v in V.items() :
    #        H[i] =  Elem({0:0, 1: v + log(Gi[i])},av[i]!=0)
    H = {i:v + log(Gi[i]) for i,v in V.items()}
    if av is None:
        logP = bioLogLogitFullChoiceSet(H,av=None,choice=choice)
    else:
        logP = bioLogLogit(H,av,choice)
    return logP



def mev(V,Gi,av,choice) :
    """ Choice probability for a MEV model.
    Args:

      V: A dictionary mapping each alternative id with the expression of the utility function.
      Gi: A dictionary mapping each alternative id with the function
       \f[
          \frac{\partial G}{\partial y_i}(e^{V_1},\ldots,e^{V_J})
      \f]
        where \f$G\f$ is the MEV generating function. If an alternative \f$i\f$ is not available, then \f$G_i = 0\f$.
      av: A dictionary mapping each alternative id with its availability condition. If None, all alternatives are considered available.
      choice: Expression producing the id of the chosen alternative.

    Return:
       Choice probability of the MEV model, given by
        \f[
          \frac{e^{V_i + \ln G_i(e^{V_1},\ldots,e^{V_J})}}{\sum_j e^{V_j + \ln G_j(e^{V_1},\ldots,e^{V_J})}}
        \f]
    """
    #    H = {}
    #    for i,v in V.items() :
    #        H[i] =  Elem({0:0, 1: v + log(Gi[i])},av[i]!=0)
    
    return exp(logmev(V,Gi,av,choice))

def logmev_selectionBias(V,Gi,av,correction,choice) :
    """ Log of choice probability for a MEV model, including the correction for endogenous sampling as proposed by <a href="http://dx.doi.org/10.1016/j.trb.2007.09.003" taret="_blank">Bierlaire, Bolduc and McFadden (2008)</a>.

    Args:
      V: A dictionary mapping each alternative id with the expression of the utility function.
     Gi: A dictionary mapping each alternative id with the function
 \f[
   \frac{\partial G}{\partial y_i}(e^{V_1},\ldots,e^{V_J})
\f]
 where \f$G\f$ is the MEV generating function.
      av: A dictionary mapping each alternative id with its availability condition. If None, all alternatives are considered available.
    correction: A dictionary mapping each alternative id with the expression of the correction. Typically, it is a value, or a parameter to be estimated.

    choice: Expression producing the id of the chosen alternative.

    Return: 
        Log of choice probability of the MEV model, given by
  \f[
    V_i + \ln G_i(e^{V_1},\ldots,e^{V_J}) - \log\left(\sum_j e^{V_j + \ln G_j(e^{V_1},\ldots,e^{V_J})}\right)
  \f]
    """
    H = {i: v + log(Gi[i]) + correction[i] for i,v in V.items()}
    logP = bioLogLogit(H,av,choice)
    return logP



def mev_selectionBias(V,Gi,av,correction,choice) :
    """ Choice probability for a MEV model, including the correction for endogenous sampling as proposed by <a href="http://dx.doi.org/10.1016/j.trb.2007.09.003" taret="_blank">Bierlaire, Bolduc and McFadden (2008)</a>.

    Args:
      V: A dictionary mapping each alternative id with the expression of the utility function.
      Gi: A dictionary mapping each alternative id with the function
      \f[
      \frac{\partial G}{\partial y_i}(e^{V_1},\ldots,e^{V_J})
      \f]
      where \f$G\f$ is the MEV generating function.
      av: A dictionary mapping each alternative id with its availability condition. If None, all alternatives are considered available.
      correction: A dictionary mapping each alternative id with the expression of the correction. Typically, it is a value, or a parameter to be estimated.
      choice: Expression producing the id of the chosen alternative.

    Return:
    Choice probability of the MEV model, given by
    \f[
    \frac{e^{V_i + \ln G_i(e^{V_1},\ldots,e^{V_J})}}{\sum_j e^{V_j + \ln G_j(e^{V_1},\ldots,e^{V_J})}}
    \f]
    """
    return exp(logmev_selectionBias(V,Gi,av,correction,choice))


def getMevForNested(V,availability,nests) :
    """ Implements the MEV generating function for the nested logit model
    Args:
      V:  A dictionary mapping each alternative id with the expression of the utility function.
      availability: A dictionary mapping each alternative id with its availability condition. If None, all alternatices are assumed to be always available.
      nests: A tuple containing as many items as nests. Each item is also a tuple containing two items
          - an object of type biogeme.expressions.Expression  representing the nest parameter,
          - a list containing the list of identifiers of the alternatives belonging to the nest.
    

    Return:
    A dictionary mapping each alternative id with the function
    \f[
    \frac{\partial G}{\partial y_i}(e^{V_1},\ldots,e^{V_J}) = e^{(\mu_m-1)V_i} \left(\sum_{i=1}^{J_m} e^{\mu_m V_i}\right)^{\frac{1}{\mu_m}-1}
    \f]
    where \f$m\f$ is the (only) nest containing alternative \f$i\f$, and
    \f$G\f$ is the MEV generating function.

    Example:
  nesta = MUA , [1,2,3]
  nestb = MUB , [4,5,6]
  nests = nesta, nestb
    """

    y = {i:exp(v) for i,v in V.items()}
    Gi = {}
    for m in nests:
        sumdict = [Elem({0:0.0,1: y[i] ** m[0]},availability[i]!=0) for i in m[1]]
        sum = bioMultSum(sumdict)
        for i in m[1]:
            Gi[i] = y[i]**(m[0]-1.0) * sum ** (1.0/m[0] - 1.0) 
    return Gi

def nested(V,availability,nests,choice) :
    """ Implements the nested logit model as a MEV model. 
    
    Args:
      V: A dictionary mapping each alternative id with the expression of the utility function.
      availability: A dictionary mapping each alternative id with its availability condition.
      nests: A tuple containing as many items as nests. Each item is also a tuple containing two items
         - an object of type biogeme.expressions.Expression  representing the nest parameter,
         - a list containing the list of identifiers of the alternatives belonging to the nest.
      choice: biogeme.expressions.Expression  producing the id of the chosen alternative.


    Return:
       Choice probability for the nested logit model, based on the derivatives of the MEV generating function produced by the function nested::getMevForNested

    Example:
    nesta = MUA , [1,2,3]
    nestb = MUB , [4,5,6]
    nests = nesta, nestb
    """
    Gi = getMevForNested(V,availability,nests)
    P = mev(V,Gi,availability,choice) 
    return P

def lognested(V,availability,nests,choice) :
    """ Implements the log of a nested logit model as a MEV model. 
    
    Args:
      V: A dictionary mapping each alternative id with the expression of the utility function.
      availability: A dictionary mapping each alternative id with its availability condition.
      nests: A tuple containing as many items as nests. Each item is also a tuple containing two items
       - an object of type biogeme.expressions.Expression  representing the nest parameter.
       - a list containing the list of identifiers of the alternatives belonging to the nest.
      choice: biogeme.expressions.Expression  producing the id of the chosen alternative.

    Return:
      Log of choice probability for the nested logit model, based on the derivatives of the MEV generating function produced by the function nested::getMevForNested

    Example:
    nesta = MUA , [1,2,3]
    nestb = MUB , [4,5,6]
    nests = nesta, nestb
    """
    Gi = getMevForNested(V,availability,nests)
    logP = logmev(V,Gi,availability,choice) 
    return logP

def nestedMevMu(V,availability,nests,choice,mu) :
    """Implements the nested logit model as a MEV model, where mu is also
    a parameter, if the user wants to test different normalization
    schemes.

    Args:
      V: A dictionary mapping each alternative id with the expression of the utility function.
      availability: A dictionary mapping each alternative id with its availability condition.
      nests: A tuple containing as many items as nests. Each item is also a tuple containing two items
         - an object of type biogeme.expressions.Expression  representing the nest parameter,
         - a list containing the list of identifiers of the alternatives belonging to the nest.
      choice: expression producing the id of the chosen alternative.
      mu: expression producing the value of the top-level scale parameter.

    Return: 
      The nested logit choice probability based on the following derivatives of the MEV generating function: 
    \f[
    \frac{\partial G}{\partial y_i}(e^{V_1},\ldots,e^{V_J}) = \mu e^{(\mu_m-1)V_i} \left(\sum_{i=1}^{J_m} e^{\mu_m V_i}\right)^{\frac{\mu}{\mu_m}-1}
    \f]
    where \f$m\f$ is the (only) nest containing alternative \f$i\f$, and
    \f$G\f$ is the MEV generating function.
    
    Example:
    nesta = MUA , [1,2,3]
    nestb = MUB , [4,5,6]
    nests = nesta, nestb
    """
    return exp(lognestedMevMu(V,availability,nests,choice,mu))

def _oldnestedMevMu(V,availability,nests,choice,mu) :
    """Implements the nested logit model as a MEV model, where mu is also
    a parameter, if the user wants to test different normalization
    schemes.

    Args:
      V: A dictionary mapping each alternative id with the expression of the utility function.
      availability: A dictionary mapping each alternative id with its availability condition.
      nests: A tuple containing as many items as nests. Each item is also a tuple containing two items
         - an object of type biogeme.expressions.Expression  representing the nest parameter,
         - a list containing the list of identifiers of the alternatives belonging to the nest.
      choice: expression producing the id of the chosen alternative.
      mu: expression producing the value of the top-level scale parameter.

    Return: 
      The nested logit choice probability based on the following derivatives of the MEV generating function: 
    \f[
    \frac{\partial G}{\partial y_i}(e^{V_1},\ldots,e^{V_J}) = \mu e^{(\mu_m-1)V_i} \left(\sum_{i=1}^{J_m} e^{\mu_m V_i}\right)^{\frac{\mu}{\mu_m}-1}
    \f]
    where \f$m\f$ is the (only) nest containing alternative \f$i\f$, and
    \f$G\f$ is the MEV generating function.
    
    Example:
    nesta = MUA , [1,2,3]
    nestb = MUB , [4,5,6]
    nests = nesta, nestb
    """
    y = {}
    for i,v in V.items() :
        y[i] = exp(v)
    
    Gi = {}
    for m in nests:
        sum = list()
        for i in m[1]:
            sum.append(Elem({0:0,1: y[i] ** m[0]},availability[i]!=0))
        for i in m[1]:
            Gi[i] = Elem({0:0,1:mu * y[i]**(m[0]-1.0) * bioMultSum(sum) ** (mu/m[0] - 1.0)},availability[i]!=0)
    P = mev(V,Gi,availability,choice) 
    return P

def lognestedMevMu(V,availability,nests,choice,mu) :
    """ Implements the log of the nested logit model as a MEV model, where mu is also a parameter, if the user wants to test different normalization schemes.

    Args:
      V: A dictionary mapping each alternative id with the expression of the utility function.
      availability: A dictionary mapping each alternative id with its availability condition.
      nests: A tuple containing as many items as nests. Each item is also a tuple containing two items
        - an object of type biogeme.expressions.Expression  representing the nest parameter,
        - a list containing the list of identifiers of the alternatives belonging to the nest.
      choice: expression producing the id of the chosen alternative.
      mu: expression producing the value of the top-level scale parameter.

    Return: 
      The nested logit choice probability based on the following derivatives of the MEV generating function: 
    \f[
    \frac{\partial G}{\partial y_i}(e^{V_1},\ldots,e^{V_J}) = \mu e^{(\mu_m-1)V_i} \left(\sum_{i=1}^{J_m} e^{\mu_m V_i}\right)^{\frac{\mu}{\mu_m}-1}
    \f]
    where \f$m\f$ is the (only) nest containing alternative \f$i\f$, and
    \f$G\f$ is the MEV generating function.
    """
    
    y = {i:exp(v) for i,v in V.items()}
    Gi = {}
    for m in nests:
        sum = [Elem({0:0,1: y[i] ** m[0]},availability[i]!=0) for i in m[1]]
        for i in m[1]:
            Gi[i] = mu * y[i]**(m[0]-1.0) * bioMultSum(sum) ** (mu/m[0] - 1.0)
    logP = logmev(V,Gi,availability,choice) 
    return logP


def cnl_avail(V,availability,nests,choice):
    """ Implements the cross-nested logit model as a MEV model. 
    Args:
       V: A dictionary mapping each alternative id with the expression of the utility function.
       availability: A dictionary mapping each alternative id with its availability condition.
       nests: A tuple containing as many items as nests. Each item is also a tuple containing two items
          - an object of type biogeme.expressions.Expression  representing the nest parameter,
          - a dictionary mapping the alternative ids with the cross-nested parameters for the corresponding nest. 
       choice: an object of type biogeme.expressions.Expression  characterizing the chosne alternative.

    Return: 
        Choice probability for the cross-nested logit model.

    Example:
    alphaA = {1: alpha1a,2: alpha2a, 3: alpha3a, 4: alpha4a,5: alpha5a, 6: alpha6a}
    alphaB = {1: alpha1b,2: alpha2b, 3: alpha3b, 4: alpha4b,5: alpha5b, 6: alpha6b}
    nesta = MUA , alphaA
    nestb = MUB , alphaB
    nests = nesta, nestb
    """
    Gi = {}
    Gidict = {}
    for k in V:
        Gidict[k] = list()
    for m in nests:
        biosumlist = list()
        for i,a in m[1].items():
            biosumlist.append(Elem({0:0,1:a**(m[0]) * exp(m[0] * (V[i]))},availability[i] != 0))
        biosum = bioMultSum(biosumlist)        
        for i,a in m[1].items():
            Gidict[i].append(Elem({0:0,1:(biosum**((1.0/m[0])-1.0)) * (a**m[0]) * exp((m[0]-1.0)*(V[i]))},availability[i] != 0))
    for k in V:
        Gi[k] = bioMultSum(Gidict[k])
    P = mev(V,Gi,availability,choice) 
    return P


def logcnl_avail(V,availability,nests,choice) :
    """ Implements the log of the cross-nested logit model as a MEV model. 
    
    Args:
     V: A dictionary mapping each alternative id with the expression of the utility function.
     availability: A dictionary mapping each alternative id with its availability condition.
     nests: A tuple containing as many items as nests. Each item is also a tuple
    containing two items:
       - an object of type biogeme.expressions.Expression  representing the nest parameter,
       - a dictionary mapping the alternative ids with the cross-nested parameters for the corresponding nest. 
      choice: Expression producing the id of the chosen alternative.

    Return: 
      Choice probability for the cross-nested logit model.

    Example:
alphaA = {1: alpha1a,
          2: alpha2a,
          3: alpha3a,
          4: alpha4a,
          5: alpha5a,
          6: alpha6a}
alphaB = {1: alpha1b,
          2: alpha2b,
          3: alpha3b,
          4: alpha4b,
          5: alpha5b,
          6: alpha6b}
nesta = MUA , alphaA
nestb = MUB , alphaB
nests = nesta, nestb
    """
    Gi = {}
    Gidict = {}
    for k in V:
        Gidict[k] = list()
    for m in nests:
        biosumlist = list()
        for i,a in m[1].items():
            biosumlist.append(Elem({0:0,1:a**(m[0]) * exp(m[0] * (V[i]))},availability[i] != 0))
        biosum = bioMultSum(biosumlist)
        for i,a in m[1].items():
            Gidict[i].append(Elem({0:0,1:(biosum**((1.0/m[0])-1.0)) * (a**m[0]) * exp((m[0]-1.0)*(V[i]))},availability[i] != 0))
    for k in V:
        Gi[k] = bioMultSum(Gidict[k])
    logP = logmev(V,Gi,availability,choice) 
    return logP


def cnlmu(V,availability,nests,choice,bmu) :
    """ Implements the cross-nested logit model as a MEV model with the homogeneity parameters is explicitly involved

    Args:
      V: A dictionary mapping each alternative id with the expression of the utility function.
      availability: A dictionary mapping each alternative id with its availability condition.
      nests: A tuple containing as many items as nests. Each item is also a tuple containing two items
        - an object of type biogeme.expressions.Expression  representing the nest parameter.
        - a dictionary mapping the alternative ids with the cross-nested parameters for the corresponding nest. 
      choice: Expression producing the id of the chosen alternative.
      bmu: Homogeneity parameter \f$\mu\f$.

    Return: 
      Choice probability for the cross-nested logit model.

    Example:
    alphaA = {1: alpha1a,
          2: alpha2a,
          3: alpha3a,
          4: alpha4a,
          5: alpha5a,
          6: alpha6a}
    alphaB = {1: alpha1b,
          2: alpha2b,
          3: alpha3b,
          4: alpha4b,
          5: alpha5b,
          6: alpha6b}
    nesta = MUA , alphaA
    nestb = MUB , alphaB
    nests = nesta, nestb
    """
    Gi = {}
    Gidict = {}
    for k in V:
        Gilist[k] = list()
    for m in nests:
        biosumlist = list()
        for i,a in m[1].items():
            biosumlist.append(Elem({0:0,1:a**(m[0]/bmu) * exp(m[0] * (V[i]))},availability[i] != 0))
        biosum = bioMultSum(biosumlist)
        for i,a in m[1].items():
            Gilist[i].append(Elem({0:0,1:bmu * (biosum**((bmu/m[0])-1.0)) * (a**(m[0]/bmu)) * exp((m[0]-1.0)*(V[i]))},availability[i] != 0))
    for k in V:
        Gi[k] = bioMultSum(Gilist[k])
    P = mev(V,Gi,availability,choice) 
    return P

def logcnlmu(V,availability,nests,choice,bmu) :
    """ Implements the log of the cross-nested logit model as a MEV model with the homogeneity parameters is explicitly involved.
    Args:
      V: A dictionary mapping each alternative id with the expression of the utility function.
      availability: A dictionary mapping each alternative id with its availability condition.
      nests: A tuple containing as many items as nests. Each item is also a tuple containing two items
        - an object of type biogeme.expressions.Expression representing the nest parameter,
        - a dictionary mapping the alternative ids with the cross-nested parameters for the corresponding nest. 
    choice: Expression producing the id of the chosen alternative.
    bmu: Homogeneity parameter \f$\mu\f$.

    Return:
      Log of choice probability for the cross-nested logit model.

    Example:
    alphaA = {1: alpha1a,
          2: alpha2a,
          3: alpha3a,
          4: alpha4a,
          5: alpha5a,
          6: alpha6a}
    alphaB = {1: alpha1b,
          2: alpha2b,
          3: alpha3b,
          4: alpha4b,
          5: alpha5b,
          6: alpha6b}
    nesta = MUA , alphaA
    nestb = MUB , alphaB
    nests = nesta, nestb
"""
    Gi = {}
    Gidict = {}
    for k in V:
        Gidict[k] = list()
    for m in nests:
        biosumlist = list() 
        for i,a in m[1].items():
            biosumlist.append(Elem({0:0,1:a**(m[0]/bmu) * exp(m[0] * (V[i]))},availability[i] != 0))
        biosum = bioMultSum(biosumlist)
        for i,a in m[1].items():
            Gidict[i].append(Elem({0:0,1:bmu * (biosum**((bmu/m[0])-1.0)) * (a**(m[0]/bmu)) * exp((m[0]-1.0)*(V[i]))},availability[i] != 0))
    for k in V:
        Gi[k] = bioMultSum(Gidict[k])
    logP = logmev(V,Gi,availability,choice) 
    return logP




import numpy as np
from scipy.optimize import *

x0 = np.array([1.3, 0.7, 0.8, 1.9, 1.2])
def rosen(x):
    """The Rosenbrock function"""
    return sum(100.0*(x[1:]-x[:-1]**2.0)**2.0 + (1-x[:-1])**2.0)

def rosen_der(x):
    xm = x[1:-1]
    xm_m1 = x[:-2]
    xm_p1 = x[2:]
    der = np.zeros_like(x)
    der[1:-1] = 200*(xm-xm_m1**2) - 400*(xm_p1 - xm**2)*xm - 2*(1-xm)
    der[0] = -400*x[0]*(x[1]-x[0]**2) - 2*(1-x[0])
    der[-1] = 200*(x[-1]-x[-2]**2)
    return der

def rosen_hess(x):
    x = np.asarray(x)
    H = np.diag(-400*x[:-1],1) - np.diag(400*x[:-1],-1)
    diagonal = np.zeros_like(x)
    diagonal[0] = 1200*x[0]**2-400*x[1]+2
    diagonal[-1] = 200
    diagonal[1:-1] = 202 + 1200*x[1:-1]**2 - 400*x[2:]
    H = H + np.diag(diagonal)
    return H

res = minimize(rosen, x0, method='Newton-CG',jac=rosen_der, hess=rosen_hess,options={'tol': 1e-8, 'disp': True})
info(minimize)




"""File 05nestedElasticities.py

:author: Michel Bierlaire, EPFL
:date: Wed Sep 11 13:41:33 2019

 We use a previously estimated nested logit model.
 Three alternatives: public transporation, car and slow modes.
 RP data.
 We calculate disaggregate and aggregate direct arc elasticities.
"""
import sys
import pandas as pd
import biogeme.database as db
import biogeme.biogeme as bio
import biogeme.models as models
import biogeme.results as res
from biogeme.expressions import Beta, DefineVariable

print("Running 05nestedElasticities.py...")

# Read the data
df = pd.read_csv("optima.dat",sep='\t')
database = db.Database("optima",df)

# The Pandas data structure is available as database.data. Use all the
# Pandas functions to invesigate the database
#print(database.data.describe())

# The following statement allows you to use the names of the variable
# as Python variable.
globals().update(database.variables)

# Exclude observations such that the chosen alternative is -1
exclude = (Choice == -1.0)
database.remove(exclude)

# Normalize the weights
sumWeight = database.data['Weight'].sum()
normalizedWeight = Weight * 1906 / 0.814484

# Calculate the number of accurences of a value in the database
numberOfMales = database.count("Gender",1)
print(f"Number of males:   {numberOfMales}")
numberOfFemales = database.count("Gender",2)
print(f"Number of females: {numberOfFemales}")
# For more complex conditions, using directly Pandas
unreportedGender = \
                   database.data[(database.data["Gender"] != 1)
                    & (database.data["Gender"] != 2)].count()["Gender"]
print(f"Unreported gender: {unreportedGender}")


# List of parameters. Their value will be set later.
ASC_CAR = Beta('ASC_CAR',0,None,None,0)
ASC_PT = Beta('ASC_PT',0,None,None,1)
ASC_SM = Beta('ASC_SM',0,None,None,0)
BETA_TIME_FULLTIME = Beta('BETA_TIME_FULLTIME',0,None,None,0)
BETA_TIME_OTHER = Beta('BETA_TIME_OTHER',0,None,None,0)
BETA_DIST_MALE = Beta('BETA_DIST_MALE',0,None,None,0)
BETA_DIST_FEMALE = Beta('BETA_DIST_FEMALE',0,None,None,0)
BETA_DIST_UNREPORTED = Beta('BETA_DIST_UNREPORTED',0,None,None,0)
BETA_COST = Beta('BETA_COST',0,None,None,0)

# Definition of variables:
# For numerical reasons, it is good practice to scale the data to
# that the values of the parameters are around 1.0.

# The following statements are designed to preprocess the data.
# It is like creating a new columns in the data file. This
# should be preferred to the statement like
# TimePT_scaled = Time_PT / 200.0
# which will cause the division to be reevaluated again and again,
# throuh the iterations. For models taking a long time to
# estimate, it may make a significant difference.

TimePT_scaled = TimePT / 200
TimeCar_scaled = TimeCar / 200
MarginalCostPT_scaled = MarginalCostPT / 10
CostCarCHF_scaled = CostCarCHF / 10
distance_km_scaled = distance_km / 5
# We investigate a scenario where the distance increases by one kilometer.
delta_dist = 1.0
distance_km_scaled_after = (distance_km + delta_dist) / 5

male = (Gender == 1)
female = (Gender == 2)
unreportedGender = (Gender == -1)

fulltime = (OccupStat == 1)
notfulltime = (OccupStat != 1)

# Definition of utility functions:
V_PT = ASC_PT + BETA_TIME_FULLTIME * TimePT_scaled * fulltime + \
       BETA_TIME_OTHER * TimePT_scaled * notfulltime + \
       BETA_COST * MarginalCostPT_scaled
V_CAR = ASC_CAR + \
        BETA_TIME_FULLTIME * TimeCar_scaled * fulltime + \
        BETA_TIME_OTHER * TimeCar_scaled * notfulltime + \
        BETA_COST * CostCarCHF_scaled
V_SM = ASC_SM + \
       BETA_DIST_MALE * distance_km_scaled * male + \
       BETA_DIST_FEMALE * distance_km_scaled * female + \
       BETA_DIST_UNREPORTED * distance_km_scaled * unreportedGender

# Utility of the slow mode whem the distance increases by 1 kilometer.
V_SM_after = ASC_SM + \
       BETA_DIST_MALE * distance_km_scaled_after * male + \
       BETA_DIST_FEMALE * distance_km_scaled_after * female + \
       BETA_DIST_UNREPORTED * distance_km_scaled_after * unreportedGender

# Associate utility functions with the numbering of alternatives
# Before
V = {0: V_PT,
     1: V_CAR,
     2: V_SM}

# After
V_after = {0: V_PT,
           1: V_CAR,
           2: V_SM_after}

# Associate the availability conditions with the alternatives.
# In this example all alternatives are available for each individual.
av = {0: 1,
      1: 1,
      2: 1}

# Definition of the nests:
# 1: nests parameter
# 2: list of alternatives

MU_NOCAR = Beta('MU_NOCAR',1.0,1.0,None,0)

CAR_NEST = 1.0 , [ 1]
NO_CAR_NEST = MU_NOCAR , [ 0, 2]
nests = CAR_NEST, NO_CAR_NEST

# The choice model is a nested logit
# Before
prob_sm = models.nested(V,av,nests,2)

# After
prob_sm_after = models.nested(V_after,av,nests,2)

# Disaggregate elasticities
direct_elas_sm_dist = \
  (prob_sm_after - prob_sm) * distance_km / (prob_sm * delta_dist)

simulate = {'weight': normalizedWeight,
            'Prob. slow modes':prob_sm,
            'direct_elas_sm_dist':direct_elas_sm_dist}

biogeme  = bio.BIOGEME(database,simulate)
biogeme.modelName = "05nestedElasticities"

# Retrieve the values of the parameters
# First, extract the names of parameters needed for the simulation
betas = biogeme.freeBetaNames

# Read the estimation results from the file
results = res.bioResults(pickleFile='01nestedEstimation.pickle')

# Extract the values that are necessary
betaValues = results.getBetaValues(betas)

# simulatedValues is a Panda dataframe with the same number of rows as
# the database, and as many columns as formulas to simulate.
simulatedValues = biogeme.simulate(betaValues)

# We calculate the elasticities
simulatedValues['Weighted prob. slow modes'] = \
  simulatedValues['weight'] * simulatedValues['Prob. slow modes']

denominator_sm = simulatedValues['Weighted prob. slow modes'].sum()

direct_elas_sm_dist = (simulatedValues['Weighted prob. slow modes']
                       * simulatedValues['direct_elas_sm_dist'] /
                       denominator_sm).sum()

print(f"Aggregate direct elasticity of slow modes wrt distance: {direct_elas_sm_dist:.3g}")




"""File 11cnl_simul.py

:author: Michel Bierlaire, EPFL
:date: Sun Sep  8 11:13:22 2019

 Example of simulation with a cross-nested logit model. 
 Three alternatives: Train, Car and Swissmetro
 Train and car are in the same nest.
 SP data
"""
import pandas as pd
import biogeme.database as db
import biogeme.biogeme as bio
import biogeme.models as models
import biogeme.results as res
from biogeme.expressions import Beta, DefineVariable, Derive

# Read the data
df = pd.read_csv("swissmetro.dat",sep='\t')
database = db.Database("swissmetro",df)

# The Pandas data structure is available as database.data. Use all the
# Pandas functions to invesigate the database
#print(database.data.describe())

# The following statement allows you to use the names of the variable
# as Python variable.
globals().update(database.variables)

# Removing some observations can be done directly using pandas.
#remove = (((database.data.PURPOSE != 1) & (database.data.PURPOSE != 3)) | (database.data.CHOICE == 0))
#database.data.drop(database.data[remove].index,inplace=True)

# Here we use the "biogeme" way for backward compatibility
exclude = (( PURPOSE != 1 ) * (  PURPOSE   !=  3  ) +  ( CHOICE == 0 )) > 0
database.remove(exclude)

# Simulation should be done with estimated value of the
# parameters. You can included them by end. Here, we prefer to set
# them to zero, and reaf the values from the file created after
# estimation.
ASC_CAR = Beta('ASC_CAR',0,None,None,0)
ASC_TRAIN = Beta('ASC_TRAIN',0,None,None,0)
ASC_SM = Beta('ASC_SM',0,None,None,1)
B_TIME = Beta('B_TIME',0,None,None,0)
B_COST = Beta('B_COST',0,None,None,0)
MU_EXISTING = Beta('MU_EXISTING',1,1,None,0)
MU_PUBLIC = Beta('MU_PUBLIC',1,1,None,0)
ALPHA_EXISTING = Beta('ALPHA_EXISTING',0.5,0,1,0)
ALPHA_PUBLIC = 1 - ALPHA_EXISTING

# Definition of new variables
SM_COST =  SM_CO   * (  GA   ==  0  ) 
TRAIN_COST =  TRAIN_CO   * (  GA   ==  0  )

# Definition of new variables: adding columns to the database 
TRAIN_TT_SCALED = TRAIN_TT / 100.0
TRAIN_COST_SCALED = DefineVariable('TRAIN_COST_SCALED',\
                                   TRAIN_COST / 100,database)
SM_TT_SCALED = SM_TT / 100.0
SM_COST_SCALED = DefineVariable('SM_COST_SCALED', SM_COST / 100,database)
CAR_TT_SCALED = CAR_TT / 100.0
CAR_CO_SCALED = DefineVariable('CAR_CO_SCALED', CAR_CO / 100,database)

# Definition of the utility functions
V1 = ASC_TRAIN + \
     B_TIME * TRAIN_TT_SCALED + \
     B_COST * TRAIN_COST_SCALED
V2 = ASC_SM + \
     B_TIME * SM_TT_SCALED + \
     B_COST * SM_COST_SCALED
V3 = ASC_CAR + \
     B_TIME * CAR_TT_SCALED + \
     B_COST * CAR_CO_SCALED

# Associate utility functions with the numbering of alternatives
V = {1: V1,
     2: V2,
     3: V3}

# Associate the availability conditions with the alternatives
CAR_AV_SP =  DefineVariable('CAR_AV_SP',CAR_AV  * (  SP   !=  0  ),database)
TRAIN_AV_SP =  DefineVariable('TRAIN_AV_SP',TRAIN_AV  * (  SP   !=  0  ),database)
av = {1: TRAIN_AV_SP,
      2: SM_AV,
      3: CAR_AV_SP}

#Definition of nests:
alpha_existing = {1: ALPHA_EXISTING,
                  2:0.0,
                  3:1.0}

alpha_public = {1: ALPHA_PUBLIC,
                2: 1.0,
                3: 0.0}

nest_existing = MU_EXISTING, alpha_existing
nest_public = MU_PUBLIC, alpha_public
nests = nest_existing, nest_public

# Instead of estimating the parameters, read the estimation
# results from the pickle file.
results = res.bioResults(pickleFile='11cnl.pickle')

print("Estimaton results: ",results.getEstimatedParameters())


# The choice model is a cross-nested logit, with availability conditions
prob1 = models.cnl_avail(V,av,nests,1)
prob2 = models.cnl_avail(V,av,nests,2)
prob3 = models.cnl_avail(V,av,nests,3)

# We calculate elasticities
genelas1 = Derive(prob1,'TRAIN_TT') * TRAIN_TT / prob1
genelas2 = Derive(prob2,'SM_TT') * SM_TT / prob2
genelas3 = Derive(prob3,'CAR_TT') * CAR_TT / prob3

# We report the probability of each alternative and the elasticities
simulate = {'Prob. train': prob1,
            'Prob. Swissmetro': prob2,
            'Prob. car':prob3,
            'Elas. 1':genelas1,
            'Elas. 2':genelas2,
            'Elas. 3':genelas3}

# Create the Biogeme object
biosim  = bio.BIOGEME(database,simulate)
biosim.modelName = "11cnl_simul"

# Perform the simulation
simresults = biosim.simulate(results.data.betaValues)
print("Simulation results")
print(simresults.describe())

print("Aggregate shares of train: ",simresults['Prob. train'].mean())





import pandas as pd
import numpy as np
import biogeme.database as db
import biogeme.biogeme as bio
import biogeme.models as models
import biogeme.distributions as dist
import unittest
from biogeme.expressions import Beta, DefineVariable, RandomVariable, Integrate, exp, log

pandas = pd.read_csv("optima.dat",sep='\t')
database = db.Database("optima",pandas)

globals().update(database.variables)

exclude = (Choice == -1.0)
database.remove(exclude)




### Variables

ScaledIncome = DefineVariable('ScaledIncome',\
                              CalculatedIncome / 1000,database)
formulaIncome = models.piecewiseFormula(ScaledIncome,[None,4,6,8,10,None],[-0.33669944071374086,0.2177481824894475,-0.6224104823788724,1.1592372281254595,-0.35530173883216076])

age_65_more = DefineVariable('age_65_more',age >= Numeric(65),database)
moreThanOneCar = DefineVariable('moreThanOneCar',NbCar > 1,database)
moreThanOneBike = DefineVariable('moreThanOneBike',NbBicy > 1,database)
individualHouse = DefineVariable('individualHouse',\
                                 HouseType == 1,database)
male = DefineVariable('male',Gender == 1,database)
haveChildren = DefineVariable('haveChildren',\
                              ((FamilSitu == 3)+(FamilSitu == 4)) > 0,database)
haveGA = DefineVariable('haveGA',GenAbST == 1,database)
highEducation = DefineVariable('highEducation', Education >= 6,database)

ASC_CAR = Beta('ASC_CAR', 0.4118057751313065,None,None,0)
ASC_SM = Beta('ASC_SM', 1.0134144912346523,None,None,0)
BETA_COST_HWH = Beta('BETA_COST_HWH', -1.7754249429871691,None,None,0)
BETA_COST_OTHER = Beta('BETA_COST_OTHER', -1.5427606901901116,None,None,0)
BETA_DIST = Beta('BETA_DIST', -4.952750818319787,None,None,0)
BETA_TIME_CAR_CL = Beta('BETA_TIME_CAR_CL', -0.1412781305192878,None,None,0)
BETA_TIME_CAR_REF = Beta('BETA_TIME_CAR_REF', -26.647254885078237,None,None,0)
BETA_TIME_PT_CL = Beta('BETA_TIME_PT_CL', -0.5048865862522101,None,None,0)
BETA_TIME_PT_REF = Beta('BETA_TIME_PT_REF', -4.883762313381118,None,None,0)
BETA_WAITING_TIME = Beta('BETA_WAITING_TIME', -0.052434936015492456,None,None,0)
coef_age_65_more = Beta('coef_age_65_more', 1.1187240773542817,None,None,0)
coef_haveChildren = Beta('coef_haveChildren', 0.18090193116052988,None,None,0)
coef_haveGA = Beta('coef_haveGA', 4.657846388649241,None,None,0)
coef_highEducation = Beta('coef_highEducation', -1.4305949806468112,None,None,0)
coef_individualHouse = Beta('coef_individualHouse', 0.1408510971738762,None,None,0)
coef_male = Beta('coef_male', -0.6817893725153135,None,None,0)
coef_moreThanOneBike = Beta('coef_moreThanOneBike', 0.3944574007669639,None,None,0)
coef_moreThanOneCar = Beta('coef_moreThanOneCar', -2.2429029134722964,None,None,0)
sigma_s = Beta('sigma_s', 3.580536793824548,None,None,0)

coef_intercept = Beta('coef_intercept',0.0,None,None,1)

### Latent variable: structural equation

# Note that the expression must be on a single line. In order to 
# write it across several lines, each line must terminate with 
# the \ symbol

omega = RandomVariable('omega')
density = dist.normalpdf(omega) 

CARLOVERS = \
coef_intercept +\
coef_age_65_more * age_65_more +\
formulaIncome+\
coef_moreThanOneCar * moreThanOneCar +\
coef_moreThanOneBike * moreThanOneBike +\
coef_individualHouse * individualHouse +\
coef_male * male +\
coef_haveChildren * haveChildren +\
coef_haveGA * haveGA +\
coef_highEducation * highEducation +\
sigma_s * omega

# Choice model


ASC_PT	 = Beta('ASC_PT',0.0,None,None,1)

TimePT_scaled  = DefineVariable('TimePT_scaled', TimePT   /  200 ,database)
TimeCar_scaled  = DefineVariable('TimeCar_scaled', TimeCar   /  200 ,database)
MarginalCostPT_scaled  = \
 DefineVariable('MarginalCostPT_scaled', MarginalCostPT   /  10 ,database)
CostCarCHF_scaled  = \
 DefineVariable('CostCarCHF_scaled', CostCarCHF   /  10 ,database)
distance_km_scaled  = \
 DefineVariable('distance_km_scaled', distance_km   /  5 ,database)
PurpHWH = DefineVariable('PurpHWH', TripPurpose == 1,database)
PurpOther = DefineVariable('PurpOther', TripPurpose != 1,database)


### DEFINITION OF UTILITY FUNCTIONS:

BETA_TIME_PT = BETA_TIME_PT_REF * \
               exp(BETA_TIME_PT_CL * CARLOVERS)

V0 = ASC_PT + \
     BETA_TIME_PT * TimePT_scaled + \
     BETA_WAITING_TIME * WaitingTimePT + \
     BETA_COST_HWH * MarginalCostPT_scaled * PurpHWH  +\
     BETA_COST_OTHER * MarginalCostPT_scaled * PurpOther

BETA_TIME_CAR = BETA_TIME_CAR_REF * \
                exp(BETA_TIME_CAR_CL * CARLOVERS)

V1 = ASC_CAR + \
      BETA_TIME_CAR * TimeCar_scaled + \
      BETA_COST_HWH * CostCarCHF_scaled * PurpHWH  + \
      BETA_COST_OTHER * CostCarCHF_scaled * PurpOther 

V2 = ASC_SM + BETA_DIST * distance_km_scaled

# Associate utility functions with the numbering of alternatives
V = {0: V0,
     1: V1,
     2: V2}

# Associate the availability conditions with the alternatives.
# In this example all alternatives are available 
# for each individual.
av = {0: 1,
      1: 1,
      2: 1}

# The choice model is a logit, conditional to 
# the value of the latent variable
condprob = models.logit(V,av,Choice)
prob = Integrate(condprob * density,'omega')
loglike = log(prob)

class test_03(unittest.TestCase):
    def testEstimation(self):
        biogeme  = bio.BIOGEME(database,loglike)
        results = biogeme.estimate()
        self.assertAlmostEqual(results.data.logLike,-1076.6955020804955,2)
        
if __name__ == '__main__':
    unittest.main()





import pandas as pd
import numpy as np
import biogeme.database as db
import biogeme.biogeme as bio
import biogeme.loglikelihood as ll
import biogeme.models as models
from biogeme.expressions import Beta, DefineVariable, Elem, bioMultSum, Numeric
import unittest

pandas = pd.read_csv("optima.dat",sep='\t')
database = db.Database("optima",pandas)

globals().update(database.variables)

exclude = (Choice == -1.0)
database.remove(exclude)


# Piecewise linear definition of income
# Note the use of the biogeme implementation of min and max, to avoid
# ambiguity with the python native functions.

ScaledIncome = DefineVariable('ScaledIncome',\
                              CalculatedIncome / 1000,database)
formulaIncome = models.piecewiseFormula(ScaledIncome,[None,4,6,8,10,None])

age_65_more = DefineVariable('age_65_more',age >= Numeric(65),database)
moreThanOneCar = DefineVariable('moreThanOneCar',NbCar > 1,database)
moreThanOneBike = DefineVariable('moreThanOneBike',NbBicy > 1,database)
individualHouse = DefineVariable('individualHouse',\
                                 HouseType == 1,database)
male = DefineVariable('male',Gender == 1,database)
haveChildren = DefineVariable('haveChildren',\
                              ((FamilSitu == 3)+(FamilSitu == 4)) > 0,database)
haveGA = DefineVariable('haveGA',GenAbST == 1,database)
highEducation = DefineVariable('highEducation', Education >= 6,database)

### Coefficients
coef_intercept = Beta('coef_intercept',0.0,None,None,0)
coef_age_65_more = Beta('coef_age_65_more',0.0,None,None,0)
coef_age_unknown = Beta('coef_age_unknown',0.0,None,None,0)
coef_haveGA = Beta('coef_haveGA',0.0,None,None,0)
coef_moreThanOneCar = \
 Beta('coef_moreThanOneCar',0.0,None,None,0)
coef_moreThanOneBike = \
 Beta('coef_moreThanOneBike',0.0,None,None,0)
coef_individualHouse = \
 Beta('coef_individualHouse',0.0,None,None,0)
coef_male = Beta('coef_male',0.0,None,None,0)
coef_haveChildren = Beta('coef_haveChildren',0.0,None,None,0)
coef_highEducation = Beta('coef_highEducation',0.0,None,None,0)

### Latent variable: structural equation

# Note that the expression must be on a single line. In order to 
# write it across several lines, each line must terminate with 
# the \ symbol

CARLOVERS = \
coef_intercept +\
coef_age_65_more * age_65_more +\
formulaIncome+\
coef_moreThanOneCar * moreThanOneCar +\
coef_moreThanOneBike * moreThanOneBike +\
coef_individualHouse * individualHouse +\
coef_male * male +\
coef_haveChildren * haveChildren +\
coef_haveGA * haveGA +\
coef_highEducation * highEducation

sigma_s = Beta('sigma_s',1,0.001,None,1)

### Measurement equations

INTER_Envir01 = Beta('INTER_Envir01',0,None,None,1)
INTER_Envir02 = Beta('INTER_Envir02',0,None,None,0)
INTER_Envir03 = Beta('INTER_Envir03',0,None,None,0)
INTER_Mobil11 = Beta('INTER_Mobil11',0,None,None,0)
INTER_Mobil14 = Beta('INTER_Mobil14',0,None,None,0)
INTER_Mobil16 = Beta('INTER_Mobil16',0,None,None,0)
INTER_Mobil17 = Beta('INTER_Mobil17',0,None,None,0)

B_Envir01_F1 = Beta('B_Envir01_F1',-1,None,None,1)
B_Envir02_F1 = Beta('B_Envir02_F1',-1,None,None,0)
B_Envir03_F1 = Beta('B_Envir03_F1',1,None,None,0)
B_Mobil11_F1 = Beta('B_Mobil11_F1',1,None,None,0)
B_Mobil14_F1 = Beta('B_Mobil14_F1',1,None,None,0)
B_Mobil16_F1 = Beta('B_Mobil16_F1',1,None,None,0)
B_Mobil17_F1 = Beta('B_Mobil17_F1',1,None,None,0)



MODEL_Envir01 = INTER_Envir01 + B_Envir01_F1 * CARLOVERS
MODEL_Envir02 = INTER_Envir02 + B_Envir02_F1 * CARLOVERS
MODEL_Envir03 = INTER_Envir03 + B_Envir03_F1 * CARLOVERS
MODEL_Mobil11 = INTER_Mobil11 + B_Mobil11_F1 * CARLOVERS
MODEL_Mobil14 = INTER_Mobil14 + B_Mobil14_F1 * CARLOVERS
MODEL_Mobil16 = INTER_Mobil16 + B_Mobil16_F1 * CARLOVERS
MODEL_Mobil17 = INTER_Mobil17 + B_Mobil17_F1 * CARLOVERS

SIGMA_STAR_Envir01 = Beta('SIGMA_STAR_Envir01',10,0.001,None,0)
SIGMA_STAR_Envir02 = Beta('SIGMA_STAR_Envir02',10,0.001,None,0)
SIGMA_STAR_Envir03 = Beta('SIGMA_STAR_Envir03',10,0.001,None,0)
SIGMA_STAR_Mobil11 = Beta('SIGMA_STAR_Mobil11',10,0.001,None,0)
SIGMA_STAR_Mobil14 = Beta('SIGMA_STAR_Mobil14',10,0.001,None,0)
SIGMA_STAR_Mobil16 = Beta('SIGMA_STAR_Mobil16',10,0.001,None,0)
SIGMA_STAR_Mobil17 = Beta('SIGMA_STAR_Mobil17',10,0.001,None,0)


F = {}
F['Envir01'] = Elem({0:0, \
 1:ll.loglikelihoodregression(Envir01,MODEL_Envir01,SIGMA_STAR_Envir01)},\
  (Envir01 > 0)*(Envir01 < 6))
F['Envir02'] = Elem({0:0, \
 1:ll.loglikelihoodregression(Envir02,MODEL_Envir02,SIGMA_STAR_Envir02)},\
  (Envir02 > 0)*(Envir02 < 6))
F['Envir03'] = Elem({0:0, \
 1:ll.loglikelihoodregression(Envir03,MODEL_Envir03,SIGMA_STAR_Envir03)},\
  (Envir03 > 0)*(Envir03 < 6))
F['Mobil11'] = Elem({0:0, \
 1:ll.loglikelihoodregression(Mobil11,MODEL_Mobil11,SIGMA_STAR_Mobil11)},\
  (Mobil11 > 0)*(Mobil11 < 6))
F['Mobil14'] = Elem({0:0, \
 1:ll.loglikelihoodregression(Mobil14,MODEL_Mobil14,SIGMA_STAR_Mobil14)},\
  (Mobil14 > 0)*(Mobil14 < 6))
F['Mobil16'] = Elem({0:0, \
 1:ll.loglikelihoodregression(Mobil16,MODEL_Mobil16,SIGMA_STAR_Mobil16)},\
  (Mobil16 > 0)*(Mobil16 < 6))
F['Mobil17'] = Elem({0:0, \
 1:ll.loglikelihoodregression(Mobil17,MODEL_Mobil17,SIGMA_STAR_Mobil17)},\
  (Mobil17 > 0)*(Mobil17 < 6))

loglike = bioMultSum(F)

class test_01(unittest.TestCase):
    def testEstimation(self):
        biogeme  = bio.BIOGEME(database,loglike)
        results = biogeme.estimate()
        self.assertAlmostEqual(results.data.logLike,-18945.44,2)
        
if __name__ == '__main__':
    unittest.main()




import biogeme.database as db
import biogeme.biogeme as bio
import biogeme.distributions as dist
import biogeme.models as models
import pandas as pd
import unittest
from biogeme.expressions import *

GROUP=Variable('GROUP')
SURVEY=Variable('SURVEY')
SP=Variable('SP')
ID=Variable('ID')
PURPOSE=Variable('PURPOSE')
FIRST=Variable('FIRST')
TICKET=Variable('TICKET')
WHO=Variable('WHO')
LUGGAGE=Variable('LUGGAGE')
AGE=Variable('AGE')
MALE=Variable('MALE')
INCOME=Variable('INCOME')
GA=Variable('GA')
ORIGIN=Variable('ORIGIN')
DEST=Variable('DEST')
TRAIN_AV=Variable('TRAIN_AV')
CAR_AV=Variable('CAR_AV')
SM_AV=Variable('SM_AV')
TRAIN_TT=Variable('TRAIN_TT')
TRAIN_CO=Variable('TRAIN_CO')
TRAIN_HE=Variable('TRAIN_HE')
SM_TT=Variable('SM_TT')
SM_CO=Variable('SM_CO')
SM_HE=Variable('SM_HE')
SM_SEATS=Variable('SM_SEATS')
CAR_TT=Variable('CAR_TT')
CAR_CO=Variable('CAR_CO')
CHOICE=Variable('CHOICE')


import biogeme.models as models


class test_biogeme(unittest.TestCase):
    def setUp(self):
        data = {'ID' : pd.Series([i+1 for i in range(21)]),'AutoTime': pd.Series([52.9,4.1,4.1,56.2,51.8,0.2,27.6,89.9,41.5,95.0,99.1,18.5,82.0,8.6,22.5,51.4,81.0,51.0,62.2,95.1,41.6]),'TransitTime': pd.Series([4.4,28.5,86.9,31.6,20.2,91.2,79.7,2.2,24.5,43.5,8.4,84.0,38.0,1.6,74.1,83.8,19.2,85.0,90.1,22.2,91.5]), 'Choice' : pd.Series([1,1,0,1,1,0,0,1,1,1,1,0,0,1,0,0,1,0,0,1,0])}
        pandas = pd.DataFrame(data)

        self.database = db.Database("akiva",pandas)

                                
    def testEstimation(self):
        ID=Variable('ID')
        AutoTime=Variable('AutoTime')
        TransitTime=Variable('TransitTime')
        Choice=Variable('Choice')

        ASC_CAR = Beta('ASC_CAR',0,None,None,0)
        B_TIME = Beta('B_TIME',0,None,None,0)

        V = {0: ASC_CAR + B_TIME * AutoTime,
             1: B_TIME * TransitTime}

        av = {0:1,1:1}

        logprob = models.loglogit(V,av,Choice)

        biogeme  = bio.BIOGEME(self.database,logprob)
        biogeme.modelName = "test"
        biogeme.generateHtml = False
        results = biogeme.estimate()
        self.assertAlmostEqual(results.data.logLike,-6.166042,2)

if __name__ == '__main__':
    unittest.main()



"""Implements some useful functions

:author: Michel Bierlaire

:date: Sun Apr 14 10:46:10 2019

"""

import numpy as np
import biogeme.messaging as msg


def findiff_g(theFunction,x):
    """Calculates the gradient of a function :math`f` using finite differences

    :param theFunction: A function object that takes a vector as an
                        argument, and returns a tuple. The first
                        element of the tuple is the value of the
                        function :math:`f`. The other elements are not
                        used.  
    :type theFunction: function

    :param x: argument of the function
    :type x: numpy.array

    :return: numpy vector, same dimension as x, containing the gradient
       calculated by finite differences.
    :rtype: numpy.array

    """
    tau = 0.0000001
    n = len(x)
    g = np.zeros(n)
    f = theFunction(x)[0]
    for i in range(n):
        xi = x.item(i)
        xp = x.copy()
        if (abs(xi) >= 1):
            s = tau * xi
        elif xi >= 0:
            s = tau
        else:
            s = -tau
        xp[i] = xi + s
        fp = theFunction(xp)[0]
        g[i] = (fp - f) / s
    return g

def findiff_H(theFunction,x):
    """Calculates the hessian of a function :math:`f` using finite differences

    :param theFunction: A function object that takes a vector as an
                        argument, and returns a tuple. The first
                        element of the tuple is the value of the
                        function :math:`f`, and the second is the
                        gradient of the function.  The other elements
                        are not used.

    :type theFunction: function

    :param x: argument of the function
    :type x: numpy.array
    
    :return: numpy matrix containing the hessian calculated by finite differences.
    :rtype: numpy.array

    """
    tau = 0.0000001
    n = len(x)
    H = np.zeros((n,n))
    g = theFunction(x)[1]
    I = np.eye(n,n)
    for i in range(n):
        xi = x.item(i)
        if (abs(xi) >= 1):
            s = tau * xi
        elif xi >= 0:
            s = tau
        else:
            s = -tau
        ei = I[i]
        gp = theFunction(x + s * ei)[1]
        H[:,i] = (gp-g).flatten() / s
    return H


def checkDerivatives(theFunction,x,names=None,logg=False):
    """Verifies the analytical derivatives of a function by comparing them with finite difference approximations.

    :param theFunction:  A function object that takes a vector as an  argument, and returns a tuple. 

          - The first element of the tuple is the value of the function :math:`f`,
          - the second is the gradient of the function,
          - the third is the hessian.

    :type theFunction: function

    :param x: arguments of the function
    :type x: numpy.array

    :param names: the names of the entries of x (for reporting).
    :type names: list(string)
    :param logg: if True, messages will be displayed. 
    :type logg: bool
    
       
    :return: tuple f,g,h,gdiff,hdiff where

          - f is the value of the function at x,
          - g is the analytical gradient,
          - h is the analytical hessian,
          - gdiff is the difference between the analytical gradient and the finite difference approximation
          - hdiff is the difference between the analytical hessian and the finite difference approximation

    :rtype: float, numpy.array,numpy.array,  numpy.array,numpy.array
    """
    f,g,h = theFunction(x)
    g_num = findiff_g(theFunction,x)
    gdiff = g - g_num
    if logg:
        logger = msg.bioMessage()
        if names is None:
            names = [f"x[{i}]" for i in range(len(x))]
        logger.detailed("x\t\tGradient\tFinDiff\t\tDifference")
        for k in range(len(gdiff)):
            logger.detailed("{:15}\t{:+E}\t{:+E}\t{:+E}".format(names[k],g.item(k),g_num.item(k),gdiff.item(k)))

    h_num = findiff_H(theFunction,x)
    hdiff = h - h_num
    if logg:
        logger.detailed("Row\t\tCol\t\tHessian\tFinDiff\t\tDifference")
        for row in range(len(hdiff)):
            for col in range(len(hdiff)):
                logger.detailed("{:15}\t{:15}\t{:+E}\t{:+E}\t{:+E}".format(names[row],names[col],h[row,col],h_num[row,col],hdiff[row,col]))
    return f,g,h,gdiff,hdiff

def getPrimeNumbers(n):
    """ Get a given number of prime numbers

    :param n: number of primes that are requested
    :type n: int

    :return: array with prime numbers
    :rtype: list(int)
    """
    total = 0
    upperBound = 100
    while total < n:
        upperBound *= 10
        primes = calculatePrimeNumbers(upperBound)
        total = len(primes)
    return primes[0:n]

def calculatePrimeNumbers(upperBound):
    """ Calculate prime numbers

    :param upperBound: prime numbers up to this value will be computed 
    :type upperBound: int

    :return: array with prime numbers
    :rtype: list(int)
    """
    mywork = [i for i in range(0,upperBound+1)]
    max = int(np.ceil(np.sqrt(float(upperBound))))
    # Remove all multiples
    for i in range(2,max+1):
        if mywork[i] != 0:
            for k in range(2*i,upperBound+1,i):
                mywork[k] = 0
    # Gather non zero entries, which are the prime numbers
    myprimes = []
    for i in range(1,upperBound+1):
        if mywork[i] != 0 and mywork[i] != 1:
            myprimes += [mywork[i]]

    return myprimes

def countNumberOfGroups(df,column):
    """ This function counts the number of groups of same value in a column. 
      For instance: 1,2,2,3,3,3,4,1,1  would give 5
    """
    df['_biogroups'] = (df[column] != df[column].shift(1)).cumsum()
    return len(df['_biogroups'].unique())
    



import pandas as pd
import numpy as np
import biogeme.database as db
import biogeme.biogeme as bio
import biogeme.models as models
import biogeme.distributions as dist
import biogeme.results as res
from biogeme.expressions import Beta, DefineVariable, RandomVariable, Integrate, log
import unittest

pandas = pd.read_csv("optima.dat",sep='\t')
database = db.Database("optima",pandas)

globals().update(database.variables)

exclude = (Choice == -1.0)
database.remove(exclude)




### Variables

ScaledIncome = DefineVariable('ScaledIncome',\
                              CalculatedIncome / 1000,database)
formulaIncome = models.piecewiseFormula(ScaledIncome,[None,4,6,8,10,None],[0.08954209471304636,-0.2209233080453265,0.2591889240542216,-0.5227805784067027,0.08430692986645968])


age_65_more = DefineVariable('age_65_more',age >= Numeric(65),database)
moreThanOneCar = DefineVariable('moreThanOneCar',NbCar > 1,database)
moreThanOneBike = DefineVariable('moreThanOneBike',NbBicy > 1,database)
individualHouse = DefineVariable('individualHouse',\
                                 HouseType == 1,database)
male = DefineVariable('male',Gender == 1,database)
haveChildren = DefineVariable('haveChildren',\
                              ((FamilSitu == 3)+(FamilSitu == 4)) > 0,database)
haveGA = DefineVariable('haveGA',GenAbST == 1,database)
highEducation = DefineVariable('highEducation', Education >= 6,database)

### Coefficients

coef_age_65_more = 0.07181055487226333
coef_haveChildren = -0.03758785262127424
coef_haveGA = -0.5785488899700475
coef_highEducation = -0.24726576867313482
coef_individualHouse = -0.08887159771570047
coef_intercept = 0.40149819890908217
coef_male = 0.0661412838697794
coef_moreThanOneBike = -0.2776091744681671
coef_moreThanOneCar = 0.5335541575826122

### Latent variable: structural equation

# Note that the expression must be on a single line. In order to 
# write it across several lines, each line must terminate with 
# the \ symbol

omega = RandomVariable('omega')
density = dist.normalpdf(omega) 
sigma_s = Beta('sigma_s',0.8625193422179722,None,None,0)


CARLOVERS = \
coef_intercept +\
coef_age_65_more * age_65_more +\
formulaIncome+\
coef_moreThanOneCar * moreThanOneCar +\
coef_moreThanOneBike * moreThanOneBike +\
coef_individualHouse * individualHouse +\
coef_male * male +\
coef_haveChildren * haveChildren +\
coef_haveGA * haveGA +\
coef_highEducation * highEducation +\
sigma_s * omega


# Choice model

ASC_CAR = Beta('ASC_CAR',0.7725067037758291,None,None,0)
ASC_SM = Beta('ASC_SM',1.8865188103480808,None,None,0)
BETA_COST_HWH = Beta('BETA_COST_HWH',-1.7800532700436242,None,None,0)
BETA_COST_OTHER = Beta('BETA_COST_OTHER',-0.8176256998217855,None,None,0)
BETA_DIST = Beta('BETA_DIST',-5.809646562001414,None,None,0)
BETA_TIME_CAR_CL = Beta('BETA_TIME_CAR_CL',-1.6818275468466484,None,None,0)
BETA_TIME_CAR_REF = Beta('BETA_TIME_CAR_REF',-17.694645513468497,None,0,0)
BETA_TIME_PT_CL = Beta('BETA_TIME_PT_CL',-1.2424575875582378,None,None,0)
BETA_TIME_PT_REF = Beta('BETA_TIME_PT_REF',-6.279351989351876,None,None,0)
BETA_WAITING_TIME = Beta('BETA_WAITING_TIME',-0.02949883465222827,None,None,0)
ASC_PT	 = Beta('ASC_PT',0,-10000,10000,1)

TimePT_scaled  = DefineVariable('TimePT_scaled', TimePT   /  200 ,database)
TimeCar_scaled  = DefineVariable('TimeCar_scaled', TimeCar   /  200 ,database)
MarginalCostPT_scaled  = \
 DefineVariable('MarginalCostPT_scaled', MarginalCostPT   /  10 ,database)
CostCarCHF_scaled  = \
 DefineVariable('CostCarCHF_scaled', CostCarCHF   /  10 ,database)
distance_km_scaled  = \
 DefineVariable('distance_km_scaled', distance_km   /  5 ,database)
PurpHWH = DefineVariable('PurpHWH', TripPurpose == 1,database)
PurpOther = DefineVariable('PurpOther', TripPurpose != 1,database)

### DEFINITION OF UTILITY FUNCTIONS:

BETA_TIME_PT = BETA_TIME_PT_REF * exp(BETA_TIME_PT_CL * CARLOVERS)

V0 = ASC_PT + \
     BETA_TIME_PT * TimePT_scaled + \
     BETA_WAITING_TIME * WaitingTimePT + \
     BETA_COST_HWH * MarginalCostPT_scaled * PurpHWH  +\
     BETA_COST_OTHER * MarginalCostPT_scaled * PurpOther

BETA_TIME_CAR = BETA_TIME_CAR_REF * exp(BETA_TIME_CAR_CL * CARLOVERS)

V1 = ASC_CAR + \
      BETA_TIME_CAR * TimeCar_scaled + \
      BETA_COST_HWH * CostCarCHF_scaled * PurpHWH  + \
      BETA_COST_OTHER * CostCarCHF_scaled * PurpOther 

V2 = ASC_SM + BETA_DIST * distance_km_scaled

# Associate utility functions with the numbering of alternatives
V = {0: V0,
     1: V1,
     2: V2}

# Associate the availability conditions with the alternatives.
# In this example all alternatives are available for each individual.
av = {0: 1,
      1: 1,
      2: 1}

# The choice model is a logit, conditional to the value of the latent variable
condprob = models.logit(V,av,Choice)
prob = Integrate(condprob * density,'omega')
loglike = log(prob)

class test_04(unittest.TestCase):
    def testEstimation(self):
        biogeme  = bio.BIOGEME(database,loglike)
        results = biogeme.estimate()
        self.assertAlmostEqual(results.data.logLike,-1092.600088343353,2)
        
if __name__ == '__main__':
    unittest.main()



